"""Contain commonly used data structures."""

from __future__ import annotations

from functools import lru_cache
from typing import NamedTuple, TypeVar, overload

_T = TypeVar("_T")


class DiBool(NamedTuple):
    """A tuple of four bools with directions."""

    top: bool = False
    right: bool = False
    bottom: bool = False
    left: bool = False

    @classmethod
    def from_value(cls, value: bool) -> DiBool:
        """Construct an instance from a single value."""
        return cls(top=value, right=value, bottom=value, left=value)


class DiInt(NamedTuple):
    """A tuple of four integers with directions."""

    top: int = 0
    right: int = 0
    bottom: int = 0
    left: int = 0

    @overload
    def __add__(self, other: tuple[int, ...], /) -> DiInt: ...
    @overload
    def __add__(self, other: tuple[_T, ...], /) -> DiInt: ...

    def __add__(self, other):
        """Add two DiInt instances together."""
        if not isinstance(other, DiInt):
            return NotImplemented

        return DiInt(
            top=self.top + other.top,
            right=self.right + other.right,
            bottom=self.bottom + other.bottom,
            left=self.left + other.left,
        )

    @classmethod
    def from_value(cls, value: int) -> DiInt:
        """Construct an instance from a single value."""
        return cls(top=value, right=value, bottom=value, left=value)


class DiStr(NamedTuple):
    """A tuple of four strings with directions."""

    top: str = ""
    right: str = ""
    bottom: str = ""
    left: str = ""

    @classmethod
    def from_value(cls, value: str) -> DiStr:
        """Construct an instance from a single value."""
        return cls(top=value, right=value, bottom=value, left=value)


class WeightedInt(NamedTuple):
    """Ainterger with an associated weight."""

    weight: int
    value: int


class WeightedDiInt(NamedTuple):
    """A tuple of four weighted integers."""

    top: WeightedInt
    right: WeightedInt
    bottom: WeightedInt
    left: WeightedInt

    # We cannot use :py:func:`functools.cached_property` here as it does not work with
    # :py:Class:`NamedTuple`s.
    @property  # type: ignore
    @lru_cache(maxsize=1)  # noqa: B019
    def unweighted(self) -> DiInt:
        """Get the padding without weights."""
        return DiInt(*(x.value for x in self))



"""Convert display data between formats."""

from __future__ import annotations

import asyncio
import inspect
import io
import logging
import threading
from hashlib import md5
from typing import TYPE_CHECKING, Generic, TypeVar
from weakref import ReferenceType, WeakValueDictionary, finalize, ref

from prompt_toolkit.data_structures import Size
from prompt_toolkit.layout.containers import WindowAlign

from euporie.core.app.current import get_app
from euporie.core.convert.registry import (
    _CONVERTOR_ROUTE_CACHE,
    _FILTER_CACHE,
    converters,
)
from euporie.core.ft.utils import to_plain_text

if TYPE_CHECKING:
    from collections.abc import Coroutine
    from pathlib import Path
    from typing import Any, ClassVar

    from PIL.Image import Image as PilImage
    from prompt_toolkit.formatted_text.base import StyleAndTextTuples
    from rich.console import ConsoleRenderable

    from euporie.core.data_structures import DiInt
    from euporie.core.style import ColorPaletteColor


T = TypeVar("T", bytes, str, "StyleAndTextTuples", "PilImage", "ConsoleRenderable")


log = logging.getLogger(__name__)


ERROR_OUTPUTS: dict[str, Any] = {
    "ansi": "(Format Conversion Error)",
    "ft": [("fg:white bg:darkred", "(Format Conversion Error)")],
}


_IO_THREAD: list[threading.Thread | None] = [None]  # dedicated conversion IO thread
_LOOP: list[asyncio.AbstractEventLoop | None] = [
    None
]  # global event loop for conversion


def get_loop() -> asyncio.AbstractEventLoop:
    """Create or return the conversion IO loop.

    The loop will be running on a separate thread.
    """
    if _LOOP[0] is None:
        loop = asyncio.new_event_loop()
        _LOOP[0] = loop
        thread = threading.Thread(
            target=loop.run_forever, name="EuporieConvertIO", daemon=True
        )
        thread.start()
        _IO_THREAD[0] = thread
    assert _LOOP[0] is not None
    # Check we are not already in the conversion event loop
    try:
        running_loop = asyncio.get_running_loop()
    except RuntimeError:
        running_loop = None
    if _LOOP[0] is running_loop:
        raise NotImplementedError(
            "Cannot call `convert` from the conversion event loop"
        )
    return _LOOP[0]


class _MetaDatum(type):
    _instances: WeakValueDictionary[tuple[Any, ...], Datum] = WeakValueDictionary()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def __call__(self, data: T, *args: Any, **kwargs: Any) -> Datum[T]:
        data_hash = Datum.get_hash(data)
        key: tuple[Any, ...] = (
            data_hash,
            *args,
            # Get defaults for non-passed kwargs
            *(
                kwargs.get(param.name, param.default)
                for param in inspect.signature(Datum.__init__).parameters.values()
                if param.default is not inspect._empty
                and param.name not in {"path", "source"}
            ),
        )
        if key in self._instances:
            return self._instances[key]
        instance = super().__call__(data, *args, **kwargs)
        self._instances[key] = instance
        return instance


class Datum(Generic[T], metaclass=_MetaDatum):
    """Class for storing and converting display data."""

    _pixel_size: tuple[int | None, int | None]
    _hash: str
    _root: ReferenceType[Datum]

    _sizes: ClassVar[dict[str, tuple[ReferenceType[Datum], Size]]] = {}

    def __init__(
        self,
        data: T,
        format: str,
        px: int | None = None,
        py: int | None = None,
        fg: str | ColorPaletteColor | None = None,
        bg: str | ColorPaletteColor | None = None,
        path: Path | None = None,
        source: Datum | None = None,
        align: WindowAlign = WindowAlign.LEFT,
    ) -> None:
        """Create a new instance of display data."""
        self.data: T = data
        self.format = format
        self.px, self.py = px, py
        self.fg = str(fg) if fg is not None else None
        self.bg = str(bg) if bg is not None else None
        self.path = path
        self.source: ReferenceType[Datum] = ref(source) if source else ref(self)
        self.align = align
        self._cell_size: tuple[int, float] | None = None
        self._conversions: dict[
            tuple[str, int | None, int | None, str | None, str | None, bool], T | None
        ] = {}
        self._queue: dict[
            tuple[str, int | None, int | None, str | None, str | None, bool],
            asyncio.Event,
        ] = {}
        self._finalizer = finalize(self, self._cleanup_datum_sizes, self.hash)
        self._finalizer.atexit = False

    def __repr__(self) -> str:
        """Return a string representation of object."""
        return f"{self.__class__.__name__}(format={self.format!r})"

    @classmethod
    def _cleanup_datum_sizes(cls, data_hash: str) -> None:
        """Remove all sizes for a given datum hash."""
        size_instances = cls._sizes
        for key, (datum_ref, _size) in list(size_instances.items()):
            datum = datum_ref()
            if not datum or datum.hash == data_hash:
                try:
                    del size_instances[key]
                except KeyError:
                    pass
            del datum

    def to_bytes(self) -> bytes:
        """Cast the data to bytes."""
        data = self.data
        if isinstance(data, str):
            return data.encode()
        elif isinstance(data, list):
            return to_plain_text(data).encode()
        elif isinstance(data, bytes):
            return data
        else:
            from PIL.Image import Image as PilImage

            if isinstance(data, PilImage):
                return data.tobytes()
            else:
                return b"Error"

    @staticmethod
    def get_hash(data: Any) -> str:
        """Calculate a hash of data."""
        if isinstance(data, bytes):
            hash_data = data
        elif isinstance(data, str):
            hash_data = data.encode()
        else:
            from PIL.Image import Image as PilImage

            if isinstance(data, PilImage):
                hash_data = data.tobytes()
            else:
                hash_data = data
        return md5(hash_data, usedforsecurity=False).hexdigest()

    @property
    def hash(self) -> str:
        """Return a hash of the `Datum`'s data."""
        try:
            return self._hash
        except AttributeError:
            value = self.get_hash(self.to_bytes())
            self._hash = value
            return value

    @property
    def root(self) -> Datum:
        """Retrieve the source datum of any conversion outputs."""
        try:
            return self._root() or self
        except AttributeError:
            root = self
            while True:
                if (source := root.source()) == root or source is None:
                    break
                else:
                    root = source
            self._root = ref(root or self)
            return root

    async def convert_async(
        self,
        to: str,
        cols: int | None = None,
        rows: int | None = None,
        fg: str | None = None,
        bg: str | None = None,
        extend: bool = True,
        bbox: DiInt | None = None,
    ) -> Any:
        """Perform conversion asynchronously, caching the result."""
        if to == self.format:
            # TODO - crop
            return self.data

        if not fg and hasattr(app := get_app(), "color_palette"):
            fg = self.fg or app.color_palette.fg.base_hex
        if not bg and hasattr(app := get_app(), "color_palette"):
            bg = self.bg or app.color_palette.bg.base_hex

        if (key_conv := (to, cols, rows, fg, bg, extend)) in self._queue:
            await self._queue[key_conv].wait()
        if key_conv in self._conversions:
            return self._conversions[key_conv]

        self._queue[key_conv] = event = asyncio.Event()

        routes = _CONVERTOR_ROUTE_CACHE[(self.format, to)]
        # log.debug(
        #     "Converting %s->'%s'@%s using routes: %s",
        #     self,
        #     to,
        #     (cols, rows),
        #     routes,
        # )
        output: T | None = None
        if routes:
            datum = self
            output = None
            for route in routes:
                for stage_a, stage_b in zip(route, route[1:]):
                    key_stage = (stage_b, cols, rows, fg, bg, extend)
                    if key_stage in self._conversions:
                        output = self._conversions[key_stage]
                    else:
                        # Find converter with lowest weight
                        for converter in sorted(
                            [
                                conv
                                for conv in converters[stage_b][stage_a]
                                if _FILTER_CACHE.get((conv,), conv.filter_)
                            ],
                            key=lambda x: x.weight,
                        ):
                            try:
                                output = await converter.func(
                                    datum, cols, rows, fg, bg, extend
                                )
                                self._conversions[key_stage] = output
                            except Exception:
                                log.debug(
                                    "Conversion step %s failed",
                                    converter,
                                    exc_info=True,
                                )
                                continue
                            else:
                                break
                        else:
                            log.warning("An error occurred during format conversion")
                            output = None
                    if output is None:
                        log.error(
                            "Failed to convert `%s`"
                            " to `%s` using route `%s` at stage `%s`",
                            self,
                            to,
                            route,
                            stage_b,
                        )
                        # Try the next route on error
                        break
                    if stage_b != route[-1]:
                        datum = Datum(
                            data=output,
                            format=stage_b,
                            px=self.px,
                            py=self.py,
                            fg=fg,
                            bg=bg,
                            path=self.path,
                            source=datum,
                        )
                else:
                    # If this route succeeded, stop trying routes
                    break

        # Crop or pad output
        # if bbox and any(bbox):

        if output is None:
            output = ERROR_OUTPUTS.get(to, "(Conversion Error)")

        event.set()
        del self._queue[key_conv]

        return output

    def _to_sync(self, coro: Coroutine) -> Any:
        """Call an async method synchronously."""
        future = asyncio.run_coroutine_threadsafe(coro, get_loop())
        return future.result()

    def convert(
        self,
        to: str,
        cols: int | None = None,
        rows: int | None = None,
        fg: str | None = None,
        bg: str | None = None,
        extend: bool = True,
    ) -> Any:
        """Convert between formats."""
        return self._to_sync(self.convert_async(to, cols, rows, fg, bg, extend))

    async def pixel_size_async(self) -> tuple[int | None, int | None]:
        """Get the dimensions of displayable data in pixels.

        Foreground and background color are set at this point if they are available, as
        data conversion outputs are cached and re-used.

        Returns:
            A tuple of the data's width in terminal columns and its aspect ratio, when
                converted to a image.

        """
        try:
            return self._pixel_size
        except AttributeError:
            pass

        px, py = self.px, self.py
        self_data = self.data
        format = self.format
        data: bytes

        while px is None or py is None:
            # Do not bother trying if the format is ANSI
            if format == "ansi":
                break

            from PIL.Image import Image as PilImage

            if isinstance(self_data, PilImage):
                px, py = self_data.size
                break

            # Decode base64 data
            if format.startswith("base64-"):
                data = await self.convert_async(to=format[7:])

            # Encode string data
            if isinstance(self_data, str):
                data = self_data.encode()

            if isinstance(self_data, bytes):
                data = self_data

            # Try using imagesize to get the size of the output
            try:
                import imagesize

                px_calc, py_calc = imagesize.get(io.BytesIO(data))
            except ValueError:
                px_calc = py_calc = -1

            if (
                format != "png"
                and px_calc <= 0
                and py_calc <= 0
                and _CONVERTOR_ROUTE_CACHE[(format, "png")]
            ):
                # Try converting to PNG on failure
                self_data = await self.convert_async(to="png")
                format = "png"
                continue

            if px is None and px_calc > 0:
                if py is not None and py_calc > 0:
                    px = int(px_calc * py / py_calc)
                else:
                    px = px_calc
            if py is None and py_calc > 0:
                if px is not None and px_calc > 0:
                    py = int(py_calc * px / px_calc)
                else:
                    py = py_calc
            break

        self._pixel_size = (px, py)
        return self._pixel_size

    def pixel_size(self) -> Any:
        """Get data dimensions synchronously."""
        return self._to_sync(self.pixel_size_async())

    async def cell_size_async(self) -> tuple[int, float]:
        """Get the cell width and aspect ratio of the displayable data.

        Returns:
            A tuple of the data's width in terminal columns and its aspect ratio, when
                converted to a image.
        """
        if self._cell_size is None:
            cols, aspect = 0, 0.0
            px, py = await self.pixel_size_async()
            if px is not None and py is not None:
                app = get_app()
                if hasattr(app, "cell_size_px"):
                    cell_px, cell_py = app.cell_size_px
                else:
                    cell_px, cell_py = 10, 20
                cols = max(1, int(px // cell_px))
                aspect = (py / cell_py) / (px / cell_px)
            self._cell_size = cols, aspect
        return self._cell_size

    def cell_size(self) -> Any:
        """Get cell width and aspect synchronously."""
        if self._cell_size is None:
            return self._to_sync(self.cell_size_async())
        return self._cell_size

    # def crop(self, bbox: DiInt) -> T:
    #     """Crop displayable data."""

    def add_size(self, size: tuple[int, int] | Size) -> str:
        """Store a size for a :py:class`Datum`."""
        sized_datum = (ref(self), Size(*size))
        key = str(hash(sized_datum))
        self._sizes[key] = sized_datum
        return key

    @classmethod
    def get_size(cls, key: str) -> tuple[Datum, Size] | None:
        """Retrieve a :py:class:`Datum` and it's size by its key."""
        if sized_datum := cls._sizes.get(key):
            datum_ref, size = sized_datum
            if datum := datum_ref():
                return datum, size
        return None



"""Define common filters."""

from __future__ import annotations

import os
from functools import cache, partial, reduce
from importlib.util import find_spec
from shutil import which
from typing import TYPE_CHECKING

# from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import (
    Condition,
    emacs_insert_mode,
    emacs_mode,
    has_completions,
    to_filter,
    vi_insert_mode,
    vi_mode,
    vi_replace_mode,
)

from euporie.core.key_binding.micro_state import MicroInputMode

if TYPE_CHECKING:
    from prompt_toolkit.filters import Filter
    from prompt_toolkit.layout.containers import Window


@cache
def command_exists(*cmds: str) -> Filter:
    """Verify a list of external commands exist on the system."""
    filters = [
        Condition(partial(lambda x: bool(which(cmd)), cmd))  # noqa: B023
        for cmd in cmds
    ]
    return reduce(lambda a, b: a & b, filters, to_filter(True))


@cache
def have_modules(*modules: str) -> Filter:
    """Verify a list of python modules are importable."""

    def try_import(module: str) -> bool:
        loader = find_spec(module)
        return loader is not None

    filters = [Condition(partial(try_import, module)) for module in modules]
    return reduce(lambda a, b: a & b, filters, to_filter(True))


# Determine if euporie is running inside a multiplexer.
in_screen = to_filter(os.environ.get("TERM", "").startswith("screen"))
in_tmux = to_filter(os.environ.get("TMUX") is not None)
in_mplex = in_tmux | in_screen


@Condition
def cursor_in_leading_ws() -> bool:
    """Determine if the cursor of the current buffer is in leading whitespace."""
    from prompt_toolkit.application.current import get_app

    before = get_app().current_buffer.document.current_line_before_cursor
    return (not before) or before.isspace()


@Condition
def cursor_at_end_of_line() -> bool:
    """Determine if the cursor of the current buffer is in leading whitespace."""
    from prompt_toolkit.application.current import get_app

    return get_app().current_buffer.document.is_cursor_at_the_end_of_line


@Condition
def has_suggestion() -> bool:
    """Determine if the current buffer can display a suggestion."""
    from prompt_toolkit.application.current import get_app

    app = get_app()
    return (
        app.current_buffer.suggestion is not None
        and len(app.current_buffer.suggestion.text) > 0
        and app.current_buffer.document.is_cursor_at_the_end_of_line
    )


@Condition
def has_tabs() -> bool:
    """Filter to show if any tabs are open in an app."""
    from euporie.core.app.current import get_app

    return bool(get_app().tabs)


@Condition
def has_dialog() -> bool:
    """Determine if a dialog is being displayed."""
    from prompt_toolkit.layout.containers import ConditionalContainer

    from euporie.core.app.current import get_app

    app = get_app()
    for dialog in app.dialogs.values():
        if isinstance(dialog.content, ConditionalContainer) and dialog.content.filter():
            return True
    return False


@Condition
def has_menus() -> bool:
    """Determine if a menu is being displayed."""
    from prompt_toolkit.layout.containers import ConditionalContainer

    from euporie.notebook.current import get_app

    app = get_app()
    for menu in app.menus.values():
        if isinstance(menu.content, ConditionalContainer) and menu.content.filter():
            return True
    return False


has_float = has_dialog | has_menus | has_completions


@Condition
def has_toolbar() -> bool:
    """Is there an active toolbar?"""
    from euporie.core.app.current import get_app
    from euporie.core.bars import BAR_BUFFERS

    return get_app().current_buffer.name in BAR_BUFFERS


@Condition
def tab_has_focus() -> bool:
    """Determine if there is a currently focused tab."""
    from euporie.core.app.current import get_app

    return get_app().tab is not None


@Condition
def kernel_tab_has_focus() -> bool:
    """Determine if there is a focused kernel tab."""
    from euporie.core.app.current import get_app
    from euporie.core.tabs.kernel import KernelTab

    return isinstance(get_app().tab, KernelTab)


@cache
def tab_type_has_focus(tab_class_path: str) -> Condition:
    """Determine if the focused tab is of a particular type."""
    from pkgutil import resolve_name

    from euporie.core.app.current import get_app

    tab_class = cache(resolve_name)

    return Condition(lambda: isinstance(get_app().tab, tab_class(tab_class_path)))


@Condition
def tab_can_save() -> bool:
    """Determine if the current tab can save it's contents."""
    from euporie.core.app.current import get_app
    from euporie.core.tabs.base import Tab

    return (
        tab := get_app().tab
    ) is not None and tab.__class__.write_file != Tab.write_file


@Condition
def pager_has_focus() -> bool:
    """Determine if there is a currently focused notebook."""
    from euporie.core.app.current import get_app

    app = get_app()
    pager = app.pager
    if pager is not None:
        return app.layout.has_focus(pager)
    return False


@Condition
def display_has_focus() -> bool:
    """Determine if there is a currently focused cell."""
    from euporie.core.app.current import get_app
    from euporie.core.widgets.display import DisplayControl

    return isinstance(get_app().layout.current_control, DisplayControl)


@Condition
def buffer_is_empty() -> bool:
    """Determine if the current buffer contains nothing."""
    from euporie.core.app.current import get_app

    return not get_app().current_buffer.text


@Condition
def buffer_is_code() -> bool:
    """Determine if the current buffer contains code."""
    from euporie.core.app.current import get_app

    return get_app().current_buffer.name == "code"


@Condition
def buffer_is_markdown() -> bool:
    """Determine if the current buffer contains markdown."""
    from euporie.core.app.current import get_app

    return get_app().current_buffer.name == "markdown"


@Condition
def micro_mode() -> bool:
    """When the micro key-bindings are active."""
    from euporie.core.app.app import ExtraEditingMode
    from euporie.core.app.current import get_app

    return get_app().editing_mode == ExtraEditingMode.MICRO


@Condition
def micro_replace_mode() -> bool:
    """Determine if the editor is in overwrite mode."""
    from euporie.core.app.current import get_app

    app = get_app()
    return app.micro_state.input_mode == MicroInputMode.REPLACE


@Condition
def micro_insert_mode() -> bool:
    """Determine if the editor is in insert mode."""
    from euporie.core.app.current import get_app

    app = get_app()
    return app.micro_state.input_mode == MicroInputMode.INSERT


@Condition
def micro_recording_macro() -> bool:
    """Determine if a micro macro is being recorded."""
    from euporie.core.app.current import get_app

    return get_app().micro_state.current_recording is not None


@Condition
def is_returnable() -> bool:
    """Determine if the current buffer has an accept handler."""
    from euporie.core.app.current import get_app

    return get_app().current_buffer.is_returnable


@Condition
def cursor_at_start_of_line() -> bool:
    """Determine if the cursor is at the start of a line."""
    from euporie.core.app.current import get_app

    return get_app().current_buffer.document.cursor_position_col == 0


@Condition
def cursor_on_first_line() -> bool:
    """Determine if the cursor is on the first line of a buffer."""
    from euporie.core.app.current import get_app

    return get_app().current_buffer.document.on_first_line


@Condition
def cursor_on_last_line() -> bool:
    """Determine if the cursor is on the last line of a buffer."""
    from euporie.core.app.current import get_app

    return get_app().current_buffer.document.on_last_line


@cache
def char_after_cursor(char: str) -> Condition:
    """Generate a condition to check for a character after the cursor."""
    from euporie.core.app.current import get_app

    return Condition(
        lambda: bool(
            (post := get_app().current_buffer.document.text_after_cursor)
            and post[0] == char
        )
    )


@Condition
def has_matching_bracket() -> bool:
    """Determine if the bracket at the cursor has a matching pair."""
    from euporie.core.app.current import get_app

    return bool(get_app().current_buffer.document.find_matching_bracket_position())


"""Determine if any binding style is in insert mode."""
insert_mode = (
    (vi_mode & vi_insert_mode)
    | (emacs_mode & emacs_insert_mode)
    | (micro_mode & micro_insert_mode)
)

"""Determine if any binding style is in replace mode."""
replace_mode = micro_replace_mode | vi_replace_mode


@Condition
def is_searching() -> bool:
    """Determine if the app is in search mode."""
    from euporie.core.app.current import get_app

    app = get_app()
    return (
        app.search_bar is not None and app.search_bar.control in app.layout.search_links
    )


@Condition
def at_end_of_buffer() -> bool:
    """Determine if the cursor is at the end of the current buffer."""
    from prompt_toolkit.application.current import get_app

    buffer = get_app().current_buffer
    return buffer.cursor_position == len(buffer.text)


@Condition
def kernel_is_python() -> bool:
    """Determine if the current notebook has a python kernel."""
    from euporie.core.app.current import get_app
    from euporie.core.tabs.kernel import KernelTab

    kernel_tab = get_app().tab
    if isinstance(kernel_tab, KernelTab):
        return kernel_tab.language == "python"
    return False


@Condition
def multiple_cells_selected() -> bool:
    """Determine if there is more than one selected cell."""
    from euporie.core.app.current import get_app
    from euporie.core.tabs.notebook import BaseNotebook

    nb = get_app().tab
    if isinstance(nb, BaseNotebook):
        return len(nb.selected_indices) > 1
    return False


def scrollable(window: Window) -> Filter:
    """Return a filter which indicates if a window is scrollable."""
    return Condition(
        lambda: (
            window.render_info is not None
            and window.render_info.content_height > window.render_info.window_height
        )
    )



"""Define the global search toolbar and related search functions."""

from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.document import Document
from prompt_toolkit.filters.app import is_searching
from prompt_toolkit.filters.base import Condition
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.key_binding.vi_state import InputMode
from prompt_toolkit.layout.controls import BufferControl, SearchBufferControl
from prompt_toolkit.search import SearchDirection
from prompt_toolkit.selection import SelectionState
from prompt_toolkit.widgets import SearchToolbar as PtkSearchToolbar

from euporie.core.app.current import get_app
from euporie.core.bars import SEARCH_BAR_BUFFER
from euporie.core.commands import add_cmd
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)

if TYPE_CHECKING:
    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text.base import AnyFormattedText
    from prompt_toolkit.layout.controls import UIControl

log = logging.getLogger(__name__)


class SearchBar(PtkSearchToolbar):
    """Search mode.

    A search toolbar with custom style and text.
    """

    def __init__(
        self,
        search_buffer: Buffer | None = None,
        vi_mode: bool = False,
        text_if_not_searching: AnyFormattedText = "",
        forward_search_prompt: AnyFormattedText = " Find: ",
        backward_search_prompt: AnyFormattedText = " Find (up): ",
        ignore_case: FilterOrBool = False,
    ) -> None:
        """Create a new search bar instance."""
        if search_buffer is None:
            search_buffer = Buffer(name=SEARCH_BAR_BUFFER)
        super().__init__(
            search_buffer=search_buffer,
            vi_mode=vi_mode,
            text_if_not_searching=text_if_not_searching,
            forward_search_prompt=to_formatted_text(
                forward_search_prompt, "class:status-field"
            ),
            backward_search_prompt=to_formatted_text(
                backward_search_prompt, "class:status-field"
            ),
        )
        self.control.key_bindings = load_registered_bindings(
            "euporie.core.bars.search.SearchBar",
            config=get_app().config,
        )
        search_state = self.control.searcher_search_state
        search_state.ignore_case = Condition(
            lambda: self.search_buffer.text.islower() or search_state.text.islower()
        )

    register_bindings(
        {
            "euporie.core.app.app:BaseApp": {
                "find": ["c-f", "f3", "f7"],
                "find-next": "c-g",
                "find-previous": "c-p",
            },
            "euporie.core.bars.search.SearchBar": {
                "accept-search": "enter",
                "stop-search": "escape",
            },
        }
    )


def find_search_control() -> tuple[SearchBufferControl | None, BufferControl | None]:
    """Find the current search buffer and buffer control."""
    current_buffer_control: BufferControl | None = None
    search_buffer_control: SearchBufferControl | None = None

    app = get_app()
    layout = app.layout
    current_control = app.layout.current_control

    if isinstance(current_control, SearchBufferControl):
        search_buffer_control = current_control

    if search_buffer_control is None and app.search_bar is not None:
        search_buffer_control = app.search_bar.control

    if search_buffer_control is not None and current_buffer_control is None:
        current_buffer_control = layout.search_links.get(search_buffer_control)

    if current_buffer_control is None and isinstance(current_control, BufferControl):
        current_buffer_control = current_control

    if (
        search_buffer_control is None
        and current_buffer_control is not None
        and current_buffer_control.search_buffer_control is not None
    ):
        search_buffer_control = current_buffer_control.search_buffer_control

    return search_buffer_control, current_buffer_control


def find_searchable_controls(
    search_buffer_control: SearchBufferControl, current_control: BufferControl | None
) -> list[BufferControl]:
    """Find list of searchable controls and the index of the next control."""
    # If a tab provides a list of buffers to search, use that. Otherwise, trawl the
    # layout for buffer controls with this as its search control
    long_list: list[UIControl]
    if tab := get_app().tab:
        try:
            long_list = list(tab.__pt_searchables__())
        except NotImplementedError:
            long_list = list(get_app().layout.find_all_controls())
    next_control_index = 0
    searchable_controls: list[BufferControl] = []
    for control in long_list:
        # Find the index of the next searchable control so we can link the search
        # control to it if the currently focused control is not searchable. This is so
        # that the next searchable control can be focused when search is completed.
        if control == current_control:
            next_control_index = len(searchable_controls)
        # Only save the control if it is searchable
        if (
            isinstance(control, BufferControl)
            and control.search_buffer_control == search_buffer_control
        ):
            # Add it to our list
            searchable_controls.append(control)
    # Cut list based on current control index
    searchable_controls = (
        searchable_controls[next_control_index:]
        + searchable_controls[:next_control_index]
    )
    return searchable_controls


def start_global_search(
    buffer_control: BufferControl | None = None,
    direction: SearchDirection = SearchDirection.FORWARD,
) -> None:
    """Start a search through all searchable `buffer_controls` in the layout."""
    search_buffer_control, current_control = find_search_control()
    if search_buffer_control is None:
        return
    searchable_controls = find_searchable_controls(
        search_buffer_control, current_control
    )

    # Stop the search if we did not find any searchable controls
    if not searchable_controls:
        return

    # If the current control is searchable, link it
    app = get_app()
    layout = app.layout
    if current_control in searchable_controls:
        assert isinstance(current_control, BufferControl)
        layout.search_links[search_buffer_control] = current_control
    else:
        # otherwise use the next after the currently selected control
        layout.search_links[search_buffer_control] = searchable_controls[0]
    # Make sure to focus the search BufferControl
    layout.focus(search_buffer_control)
    # If we're in Vi mode, make sure to go into insert mode.
    app.vi_state.input_mode = InputMode.INSERT


@add_cmd(menu_title="Find")
def find() -> None:
    """Enter search mode."""
    start_global_search(direction=SearchDirection.FORWARD)


def find_prev_next(direction: SearchDirection) -> None:
    """Find the previous or next search match."""
    if is_searching():
        accept_search()

    search_buffer_control, current_control = find_search_control()
    if search_buffer_control is None:
        return
    searchable_controls = find_searchable_controls(
        search_buffer_control, current_control
    )

    if direction == SearchDirection.BACKWARD:
        searchable_controls = searchable_controls[:1] + searchable_controls[1:][::-1]

    # Search over all searchable buffers
    for i, control in enumerate(searchable_controls):
        # Update search_state.
        search_state = control.search_state
        search_state.direction = direction
        # Apply search to buffer
        buffer = control.buffer

        search_result: tuple[int, int] | None = None

        # If we are searching history, use the PTK buffer search implementation
        if buffer.enable_history_search():
            search_result = buffer._search(search_state)

        # Otherwise, only search the buffer's current "working line"
        else:
            document = buffer.document
            # If have move to the next buffer, set the cursor position for the start of
            # the search to the start or the end of the text, depending on if we are
            # searching forwards or backwards
            if i > 0:
                if direction == SearchDirection.FORWARD:
                    document = Document(document.text, 0)
                else:
                    document = Document(document.text, len(document.text))

            text = search_state.text
            ignore_case = search_state.ignore_case()

            if direction == SearchDirection.FORWARD:
                # Try find at the current input.
                new_index = document.find(
                    text,
                    # If we have moved to the next buffer, include the current position
                    # which will be the start of the document text
                    include_current_position=i > 0,
                    ignore_case=ignore_case,
                )
                if new_index is not None:
                    search_result = (
                        buffer.working_index,
                        document.cursor_position + new_index,
                    )
            else:
                # Try find at the current input.
                new_index = document.find_backwards(text, ignore_case=ignore_case)
                if new_index is not None:
                    search_result = (
                        buffer.working_index,
                        document.cursor_position + new_index,
                    )

        if search_result is not None:
            working_index, cursor_position = search_result
            buffer.working_index = working_index
            buffer.cursor_position = cursor_position
            # Set SelectionState
            buffer.selection_state = SelectionState(
                buffer.cursor_position + len(search_state.text)
            )
            buffer.selection_state.enter_shift_mode()

            # Trigger a cursor position changed event on this buffer
            buffer._cursor_position_changed()

            break


@add_cmd()
def find_next() -> None:
    """Find the next search match."""
    find_prev_next(SearchDirection.FORWARD)


@add_cmd()
def find_previous() -> None:
    """Find the previous search match."""
    find_prev_next(SearchDirection.BACKWARD)


@add_cmd(
    filter=is_searching,
)
def stop_search() -> None:
    """Abort the search."""
    layout = get_app().layout
    buffer_control = layout.search_target_buffer_control
    if buffer_control is None:
        return
    search_buffer_control = buffer_control.search_buffer_control
    # Focus the previous control
    layout.focus(layout.previous_control)
    # Close the search toolbar
    if search_buffer_control is not None:
        del layout.search_links[search_buffer_control]
        # Reset content of search control.
        search_buffer_control.buffer.reset()
    # Redraw everything
    get_app().refresh()


@add_cmd(
    name="accept-search",
    filter=is_searching,
)
def accept_search() -> None:
    """Accept the search input."""
    layout = get_app().layout
    search_buffer_control = layout.current_control
    if not isinstance(search_buffer_control, BufferControl):
        log.debug("Current control not a buffer control")
        return
    for control in layout.find_all_controls():
        if (
            isinstance(control, BufferControl)
            and control.search_buffer_control == search_buffer_control
        ):
            search_state = control.search_state
            # Update search state.
            if search_buffer_control.buffer.text:
                search_state.text = search_buffer_control.buffer.text
            # Apply search.
            control.buffer.apply_search(
                search_state, include_current_position=True, count=1
            )

    # Set selection on target control
    buffer_control = layout.search_target_buffer_control
    if buffer_control and control.is_focusable():
        buffer = buffer_control.buffer
        buffer.selection_state = SelectionState(
            buffer.cursor_position + len(search_state.text)
        )
        buffer.selection_state.enter_shift_mode()

    # Add query to history of search line.
    search_buffer_control.buffer.append_to_history()
    # Stop the search
    stop_search()


@add_cmd()
def _replace_all(find_str: str, replace_str: str) -> None:
    """Find and replace text in all searchable buffers.

    Args:
        find_str: String pattern to find (will be converted to regex)
        replace_str: Replacement string
    """
    # Convert find string to regex pattern
    pattern = re.compile(find_str)

    # Get searchable controls
    search_buffer_control, current_control = find_search_control()
    if search_buffer_control is None:
        return
    searchable_controls = find_searchable_controls(
        search_buffer_control, current_control
    )

    # Apply replacements to each buffer
    for control in searchable_controls:
        if isinstance(control, BufferControl):
            buffer = control.buffer
            text = buffer.text
            new_text = pattern.sub(replace_str, text)
            if new_text != text:
                buffer.text = new_text
                buffer.on_text_changed()



"""Contain functions which convert data to PIL format."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from euporie.core.convert.registry import register
from euporie.core.filters import have_modules

if TYPE_CHECKING:
    from PIL.Image import Image as PilImage

    from euporie.core.convert.datum import Datum


log = logging.getLogger(__name__)


def set_background(image: PilImage, bg: str | None = None) -> PilImage:
    """Remove the alpha channel from an image and set the background colour."""
    from PIL import Image

    if image.mode in ("RGBA", "LA") or (
        image.mode == "P" and "transparency" in image.info
    ):
        alpha = image.convert("RGBA").getchannel("A")
        bg_img = Image.new("RGBA", image.size, bg or "#000")
        bg_img.paste(image, mask=alpha)
        image = bg_img
    return image.convert("P", palette=Image.Palette.ADAPTIVE, colors=16).convert(
        "RGB", palette=Image.Palette.ADAPTIVE, colors=16
    )


@register(
    from_=("png", "jpeg", "gif"),
    to="pil",
    filter_=have_modules("PIL"),
)
async def png_to_pil_py(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> PilImage:
    """Convert PNG to a pillow image using :py:mod:`PIL`."""
    import io

    from PIL import Image

    try:
        image = Image.open(io.BytesIO(datum.data))
        image.load()
    except OSError:
        log.error("Could not load image.")
        return Image.new(mode="P", size=(1, 1))
    else:
        return image


'''
def crop(data: PilImage, bbox: DiInt) -> PilImage:
    """Ctop a pillow image."""
    import io

    image = data.convert(
        to="pil",
        cols=full_width,
        rows=full_height,
    )
    if image is not None:
        cell_size_x, cell_size_y = self.app.cell_size_px
        # Downscale image to fit target region for precise cropping
        image.thumbnail((full_width * cell_size_x, full_height * cell_size_y))
        image = image.crop(
            (
                self.bbox.left * cell_size_x,  # left
                self.bbox.top * cell_size_y,  # top
                (self.bbox.left + cols) * cell_size_x,  # right
                (self.bbox.top + rows) * cell_size_y,  # bottom
            )
        )
        with io.BytesIO() as output:
            image.save(output, format="PNG")
            Datum(data=output.getvalue(), format="png")
'''



"""Define custom inputs and outputs, and related methods."""

from __future__ import annotations

import logging
import re
from base64 import b64encode
from functools import lru_cache
from typing import TYPE_CHECKING, cast

from prompt_toolkit.input import vt100_parser
from prompt_toolkit.input.ansi_escape_sequences import ANSI_SEQUENCES
from prompt_toolkit.input.base import DummyInput, _dummy_context_manager
from prompt_toolkit.output.color_depth import ColorDepth
from prompt_toolkit.output.vt100 import Vt100_Output as PtkVt100_Output

from euporie.core.app.current import get_app
from euporie.core.filters import in_screen, in_tmux

if TYPE_CHECKING:
    from contextlib import AbstractContextManager
    from typing import IO, Any, Callable, TextIO

    from prompt_toolkit.keys import Keys

    from euporie.core.config import Config

log = logging.getLogger(__name__)

COLOR_DEPTHS = {
    1: ColorDepth.DEPTH_1_BIT,
    4: ColorDepth.DEPTH_4_BIT,
    8: ColorDepth.DEPTH_8_BIT,
    24: ColorDepth.DEPTH_24_BIT,
}


@lru_cache
def _have_termios_tty_fcntl() -> bool:
    try:
        import fcntl  # noqa F401
        import termios  # noqa F401
        import tty  # noqa F401
    except ModuleNotFoundError:
        return False
    else:
        return True


def _tiocgwinsz() -> tuple[int, int, int, int]:
    """Get the size and pixel dimensions of the terminal with `termios`."""
    import array

    output = array.array("H", [0, 0, 0, 0])
    if _have_termios_tty_fcntl():
        import fcntl
        import termios

        try:
            fcntl.ioctl(1, termios.TIOCGWINSZ, output)
        except OSError:
            pass
    rows, cols, xpixels, ypixels = output
    return rows, cols, xpixels, ypixels


def passthrough(cmd: str, config: Config | None = None) -> str:
    """Wrap an escape sequence for terminal passthrough."""
    config = config or get_app().config
    if config.multiplexer_passthrough:
        if in_tmux():
            cmd = cmd.replace("\x1b", "\x1b\x1b")
            cmd = f"\x1bPtmux;{cmd}\x1b\\"
        elif in_screen():
            # Screen limits escape sequences to 768 bytes, so we have to chunk it
            cmd = "".join(
                f"\x1bP{cmd[i : i + 764]}\x1b\\" for i in range(0, len(cmd), 764)
            )
    return cmd


class _IsPrefixOfLongerMatchCache(vt100_parser._IsPrefixOfLongerMatchCache):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        # Pattern for any ANSI escape sequence
        self._response_prefix_re = re.compile(
            r"""^\x1b(
                \][^\\\x07]*  # Operating System Commands
                |
                _[^\\]*  # Application Program Command
                |
                \[\?[\d;]*  # Primary device attribute responses
                |
                P[ -~]*(\x1b|x1b\\)?
            )\Z""",
            re.VERBOSE,
        )
        # Generate prefix matches for all known ansi escape sequences
        # This is faster than PTK's method
        self._ansi_sequence_prefixes = {
            seq[:i] for seq in ANSI_SEQUENCES for i in range(len(seq))
        }

    def __missing__(self, prefix: str) -> bool:
        """Check if the response might match an OSC or APC code, or DA response."""
        result = bool(
            # (hard coded) If this could be a prefix of a CPR response, return True.
            vt100_parser._cpr_response_prefix_re.match(prefix)
            # True if this could be a mouse event sequence
            or vt100_parser._mouse_event_prefix_re.match(prefix)
            # True if this could be the prefix of an expected escape sequence
            or prefix in self._ansi_sequence_prefixes
            # If this could be a prefix of any other escape sequence, return True
            or self._response_prefix_re.match(prefix)
        )
        self[prefix] = result
        return result


# Monkey patch the prefix cache
vt100_parser._IS_PREFIX_OF_LONGER_MATCH_CACHE = _IsPrefixOfLongerMatchCache()


class Vt100Parser(vt100_parser.Vt100Parser):
    """A Vt100Parser which checks input against additional key patterns."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create a new VT100 parser."""
        from euporie.core.keys import MoreKeys

        super().__init__(*args, **kwargs)
        self.patterns: dict[Keys | MoreKeys, re.Pattern] = {
            MoreKeys.ColorsResponse: re.compile(
                r"^\x1b\](?P<c>(\d+;)?\d+)+;rgb:"
                r"(?P<r>[0-9A-Fa-f]{2,4})\/"
                r"(?P<g>[0-9A-Fa-f]{2,4})\/"
                r"(?P<b>[0-9A-Fa-f]{2,4})"
                # Allow BEL or ST as terminator
                r"(?:\x1b\\|\x9c|\x07)"
            ),
            MoreKeys.PixelSizeResponse: re.compile(r"^\x1b\[4;(?P<y>\d+);(?P<x>\d+)t"),
            MoreKeys.KittyGraphicsStatusResponse: re.compile(
                r"^\x1b_Gi=(4294967295|0);(?P<status>OK)\x1b\\"
            ),
            MoreKeys.SixelGraphicsStatusResponse: re.compile(
                r"^\x1b\[\?(?:\d+;)*(?P<sixel>4)(?:;\d+)*c"
            ),
            MoreKeys.ItermGraphicsStatusResponse: re.compile(
                r"^\x1bP>\|(?P<term>[^\x1b]+)\x1b\\"
            ),
            MoreKeys.SgrPixelStatusResponse: re.compile(r"^\x1b\[\?1016;(?P<Pm>\d)\$"),
            MoreKeys.ClipboardDataResponse: re.compile(
                r"^\x1b\]52;(?:c|p)?;(?P<data>[A-Za-z0-9+/=]+)\x1b\\"
            ),
        }

    def _get_match(self, prefix: str) -> None | Keys | tuple[Keys, ...]:
        """Check for additional key matches first."""
        for key, pattern in self.patterns.items():
            if pattern.match(prefix):
                return cast("Keys", key)

        return super()._get_match(prefix)


class IgnoredInput(DummyInput):
    """An input which ignores input but does not immediately close the app."""

    def attach(
        self, input_ready_callback: Callable[[], None]
    ) -> AbstractContextManager[None]:
        """Do not call the callback, so the input is never closed."""
        return _dummy_context_manager()


class Vt100_Output(PtkVt100_Output):
    """A Vt100 output which enables SGR pixel mouse positioning."""

    def enable_sgr_pixel(self) -> None:
        """Enable SGR-pixel mouse positioning."""
        self.write_raw("\x1b[?1016h")

    def disable_sgr_pixel(self) -> None:
        """Disable SGR-pixel mouse positioning."""
        self.write_raw("\x1b[?1016l")

    def enable_private_sixel_colors(self) -> None:
        """Enable private color registers for sixel graphics."""
        self.write_raw("\x1b[1070h")

    def disable_private_sixel_colors(self) -> None:
        """Disable private color registers for sixel graphics."""
        self.write_raw("\x1b[1070l")

    def enable_extended_keys(self) -> None:
        """Request extended keys."""
        # xterm
        self.write_raw("\x1b[>4;1m")
        # kitty
        self.write_raw("\x1b[>1u")

    def disable_extended_keys(self) -> None:
        """Disable extended keys."""
        # xterm
        self.write_raw("\x1b[>4;0m")
        # kitty
        self.write_raw("\x1b[<1u")

    def set_clipboard(self, text: str) -> None:
        """Set clipboard data using OSC-52."""
        b64data = b64encode(text.encode()).decode()
        self.write_raw(f"\x1b]52;c;{b64data}\x1b\\")

    def get_clipboard(self) -> None:
        """Get clipboard contents using OSC-52."""
        self.write_raw("\x1b]52;c;?\x1b\\")

    def get_colors(self) -> None:
        """Query terminal colors."""
        self.write_raw(
            passthrough(
                ("\x1b]10;?\x1b\\\x1b]11;?\x1b\\")
                + "".join(f"\x1b]4;{i};?\x1b\\" for i in range(16))
            )
        )

    def get_pixel_size(self) -> None:
        """Check the terminal's dimensions in pixels."""
        self.write_raw("\x1b[14t")

    def get_kitty_graphics_status(self) -> None:
        """Query terminal to check for kitty graphics support."""
        self.write_raw(
            "\x1b[s"
            + passthrough("\x1b_Gi=4294967295,s=1,v=1,a=q,t=d,f=24;aaaa\x1b\\")
            + "\x1b[u\x1b[2k"
        )

    def get_sixel_graphics_status(self) -> None:
        """Query terminal for sixel graphics support."""
        self.write_raw(passthrough("\x1b[c"))

    def get_iterm_graphics_status(self) -> None:
        """Query terminal for iTerm graphics support."""
        self.write_raw(passthrough("\x1b[>q"))

    def get_sgr_pixel_status(self) -> None:
        """Query terminal to check for Pixel SGR support."""
        # Enable, check, disable
        self.write_raw("\x1b[?1016h\x1b[?1016$p\x1b[?1016l")

    def get_csiu_status(self) -> None:
        """Query terminal to check for CSI-u support."""
        self.write_raw("\x1b[?u")


class PseudoTTY:
    """Make an output stream look like a TTY."""

    fake_tty = True

    def __init__(self, underlying: IO[str] | TextIO, isatty: bool = True) -> None:
        """Wrap an underlying output stream.

        Args:
            underlying: The underlying output stream
            isatty: The value to return from :py:method:`PseudoTTY.isatty`.

        """
        self._underlying = underlying
        self._isatty = isatty

    def isatty(self) -> bool:
        """Determine if the stream is interpreted as a TTY."""
        return self._isatty

    def __getattr__(self, name: str) -> Any:
        """Return an attribute of the wrappeed stream."""
        return getattr(self._underlying, name)


def edit_in_editor(filename: str, line_number: int = 0) -> None:
    """Suspend the current app and edit a file in an external editor."""
    import os
    import shlex
    import subprocess

    from prompt_toolkit.application.run_in_terminal import run_in_terminal

    def _open_file_in_editor(filename: str) -> None:
        """Call editor executable."""
        # If the 'VISUAL' or 'EDITOR' environment variable has been set, use that.
        # Otherwise, fall back to the first available editor that we can find.
        for editor in [
            os.environ.get("VISUAL"),
            os.environ.get("EDITOR"),
            "editor",
            "micro",
            "nano",
            "pico",
            "vi",
            "emacs",
        ]:
            if editor:
                try:
                    # Use 'shlex.split()' because $VISUAL can contain spaces and quotes
                    subprocess.call([*shlex.split(editor), filename])
                    return
                except OSError:
                    # Executable does not exist, try the next one.
                    pass

    async def run() -> None:
        # Open in editor
        # (We need to use `run_in_terminal`, because not all editors go to
        # the alternate screen buffer, and some could influence the cursor
        # position)
        await run_in_terminal(lambda: _open_file_in_editor(filename), in_executor=True)

    get_app().create_background_task(run())



"""Contain function which convert data to SVG format."""

from __future__ import annotations

from typing import TYPE_CHECKING

from euporie.core.convert.registry import register
from euporie.core.filters import have_modules

if TYPE_CHECKING:
    from euporie.core.convert.datum import Datum


@register(from_="latex", to="svg", filter_=have_modules("ziamath"))
async def latex_to_svg_py_ziamath(
    datum: Datum,
    cols: int | None = None,
    rows: int | None = None,
    fg: str | None = None,
    bg: str | None = None,
    extend: bool = True,
) -> str:
    """Convert LaTeX to SVG using :py:mod:`ziamath`."""
    import ziamath as zm

    data = datum.data.strip()
    if not data.startswith("$"):
        data = f"$${data}$$"
    latex = zm.Text(data, color=fg, size=12)
    return latex.svg()



"""Launch the euporie application."""


def main() -> "None":
    """Call the main entrypoint to the application."""
    import sys

    from euporie.core import __main__

    __main__.main(str(sys.modules[__name__].__package__).rpartition(".")[2])


if __name__ == "__main__":
    main()



"""Add additional keys to the prompt_toolkit vi key-bindings."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from prompt_toolkit.buffer import indent, unindent
from prompt_toolkit.filters.app import vi_insert_mode
from prompt_toolkit.key_binding.bindings.vi import (
    load_vi_bindings as load_ptk_vi_bindings,
)

from euporie.core.filters import cursor_in_leading_ws

if TYPE_CHECKING:
    from prompt_toolkit.key_binding.key_bindings import (
        ConditionalKeyBindings,
        KeyBindings,
        KeyBindingsBase,
    )
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent


def load_vi_bindings() -> KeyBindingsBase:
    """Load vi keybindings from PTK, adding additional bindings."""
    # We know the type of the vi bindings
    vi_bindings = cast(
        "KeyBindings",
        cast("ConditionalKeyBindings", load_ptk_vi_bindings()).key_bindings,
    )
    handle = vi_bindings.add

    @handle("c-i", filter=vi_insert_mode & cursor_in_leading_ws)
    def _indent(event: KeyPressEvent) -> None:
        """Indent lines."""
        buffer = event.current_buffer
        current_row = buffer.document.cursor_position_row
        indent(buffer, current_row, current_row + event.arg)

    @handle("s-tab", filter=vi_insert_mode & cursor_in_leading_ws)
    def _unindent(event: KeyPressEvent) -> None:
        """Unindent lines."""
        current_row = event.current_buffer.document.cursor_position_row
        unindent(event.current_buffer, current_row, current_row + event.arg)

    return vi_bindings



"""Define a base class for configurable apps."""

from __future__ import annotations

from abc import ABC
from inspect import isabstract
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, ClassVar

    from euporie.core.config import Config


class ConfigurableApp(ABC):
    """An application with configuration."""

    name: str | None = None
    config: Config
    _config_defaults: ClassVar[dict[str, Any]] = {"log_level_stdout": "error"}

    def __init_subclass__(cls) -> None:
        """Create a config instance for each non-abstract subclass."""
        if not isabstract(cls):
            from euporie.core.config import Config

            # Load settings
            cls.load_settings()
            cls.config = Config(
                app=cls.name,
                _help=cls.__doc__ or "",
                **cls._config_defaults,
            )

    @classmethod
    def load_settings(cls) -> None:
        """Load all known settings for this class."""
        from euporie.core.utils import import_submodules, root_module

        roots = {
            root_module(base.__module__)
            for base in cls.__mro__
            if base.__module__.startswith("euporie.")
        }
        for root in roots:
            import_submodules(root, ("_settings", "_commands"))

    @classmethod
    def launch(cls) -> None:
        """Launch the app."""
        cls.config.load()



"""Define widget for defining layouts."""

from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from functools import lru_cache, partial
from typing import TYPE_CHECKING, ClassVar, NamedTuple, cast

from prompt_toolkit.application.current import get_app
from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.filters import Condition, to_filter
from prompt_toolkit.formatted_text.base import to_formatted_text
from prompt_toolkit.formatted_text.utils import fragment_list_width
from prompt_toolkit.key_binding.key_bindings import KeyBindings
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    DynamicContainer,
    to_container,
)
from prompt_toolkit.layout.controls import (
    FormattedTextControl,
    GetLinePrefixCallable,
    UIContent,
    UIControl,
)
from prompt_toolkit.layout.dimension import Dimension as D
from prompt_toolkit.layout.dimension import to_dimension
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.mouse_events import MouseButton, MouseEventType
from prompt_toolkit.utils import Event

from euporie.core.border import OutsetGrid
from euporie.core.data_structures import DiBool
from euporie.core.ft.utils import truncate
from euporie.core.layout.containers import HSplit, VSplit, Window
from euporie.core.widgets.decor import Border

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Callable

    from prompt_toolkit.filters import FilterOrBool
    from prompt_toolkit.formatted_text.base import (
        AnyFormattedText,
        OneStyleAndTextTuple,
        StyleAndTextTuples,
    )
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import AnyContainer, Container, _Split
    from prompt_toolkit.layout.dimension import AnyDimension
    from prompt_toolkit.mouse_events import MouseEvent

    from euporie.core.border import GridStyle

log = logging.getLogger(__name__)


class Box:
    """Add padding around a container.

    This also makes sure that the parent can provide more space than required by
    the child. This is very useful when wrapping a small element with a fixed
    size into a ``VSplit`` or ``HSplit`` object. The ``HSplit`` and ``VSplit``
    try to make sure to adapt respectively the width and height, possibly
    shrinking other elements. Wrapping something in a ``Box`` makes it flexible.

    Args:
        body: Another container object.
        padding: The margin to be used around the body. This can be
        overridden by `padding_left`, padding_right`, `padding_top` and
            `padding_bottom`.
        style: A style string.
        char: Character to be used for filling the space around the body.
            (This is supposed to be a character with a terminal width of 1.)
    """

    def __init__(
        self,
        body: AnyContainer,
        padding: AnyDimension = None,
        padding_left: AnyDimension = None,
        padding_right: AnyDimension = None,
        padding_top: AnyDimension = None,
        padding_bottom: AnyDimension = None,
        width: AnyDimension = None,
        height: AnyDimension = None,
        style: str = "",
        char: None | str | Callable[[], str] = None,
        modal: bool = False,
        key_bindings: KeyBindingsBase | None = None,
    ) -> None:
        """Initialize this widget."""
        if padding is None:
            padding = D(preferred=0)

        def get(value: AnyDimension) -> D:
            if value is None:
                value = padding
            return to_dimension(value)

        self.padding_left = get(padding_left)
        self.padding_right = get(padding_right)
        self.padding_top = get(padding_top)
        self.padding_bottom = get(padding_bottom)
        self.body = body

        self.container = HSplit(
            [
                Window(height=self.padding_top, char=char),
                VSplit(
                    [
                        Window(width=self.padding_left, char=char),
                        body,
                        Window(width=self.padding_right, char=char),
                    ]
                ),
                Window(height=self.padding_bottom, char=char),
            ],
            width=width,
            height=height,
            style=style,
            modal=modal,
            key_bindings=key_bindings,
        )

    def __pt_container__(self) -> Container:
        """Return the main container for this widget."""
        return self.container


class ConditionalSplit:
    """A split container where the orientation depends on a filter."""

    def __init__(self, vertical: FilterOrBool, *args: Any, **kwargs: Any) -> None:
        """Create a new conditional split container.

        Args:
            vertical: A filter which determines if the container should be displayed vertically
            args: Positional arguments to pass to the split container
            kwargs: Key-word arguments to pass to the split container

        """
        self.vertical = to_filter(vertical)
        self.args = args
        self.kwargs = kwargs
        self._cache: SimpleCache = SimpleCache(maxsize=2)

    def load_container(self, vertical: bool) -> _Split:
        """Load the container."""
        if vertical:
            return HSplit(*self.args, **self.kwargs)
        else:
            return VSplit(*self.args, **self.kwargs)

    def container(self) -> _Split:
        """Return the container for the current orientation."""
        vertical = self.vertical()
        return self._cache.get(vertical, partial(self.load_container, vertical))

    def __pt_container__(self) -> AnyContainer:
        """Return a dynamic container."""
        return DynamicContainer(self.container)


class ReferencedSplit:
    """A split container which maintains a reference to it's children."""

    def __init__(
        self,
        split: type[_Split],
        children: Sequence[AnyContainer],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Create a new instance of the split container.

        Args:
            split: The split container class (:class:`HSplit` or :class:`VSplit`)
            children: A list of child containers
            args: Positional arguments to pass to the split container
            kwargs: Key-word arguments to pass to the split container

        """
        self.container = split([], *args, **kwargs)
        self.children = list(children)

    @property
    def children(self) -> list[AnyContainer]:
        """Convert the referenced children to containers."""
        return self._children

    @children.setter
    def children(self, children: list[AnyContainer]) -> None:
        """Set the containers children."""
        self._children = children
        self.container.children = [to_container(x) for x in self._children]

    def __pt_container__(self) -> Container:
        """Return the child container."""
        return self.container


class TabBarTab(NamedTuple):
    """A class representing a tab and its callbacks."""

    title: AnyFormattedText
    on_activate: Callable[[], NotImplementedOrNone]
    on_deactivate: Callable[[], NotImplementedOrNone] | None = None
    on_close: Callable[[], NotImplementedOrNone] | None = None
    closeable: bool = False

    def __hash__(self) -> int:
        """Hash the Tab based on current title value."""
        return hash(tuple(to_formatted_text(self.title))) * hash(
            (self.on_activate, self.on_deactivate, self.on_close, self.closeable)
        )


class TabBarControl(UIControl):
    """A control which shows a tab bar."""

    char_scroll_left = "◀"
    char_scroll_right = "▶"
    char_close: ClassVar[str] = "✖"

    def __init__(
        self,
        tabs: Sequence[TabBarTab] | Callable[[], Sequence[TabBarTab]],
        active: int | Callable[[], int],
        spacing: int = 1,
        max_title_width: int = 30,
        grid: GridStyle = OutsetGrid,
    ) -> None:
        """Create a new tab bar instance.

        Args:
            tabs: A list to tuples describing the tab title and the callback to run
                when the tab is activated.
            active: The index of the currently active tab
            spacing: The number of characters between the tabs
            max_title_width: The maximum width of the title to display
            grid: The grid style to use for drawing borders

        """
        self._tabs = tabs
        self.spacing = spacing
        self.max_title_width = max_title_width
        self._active = active
        self._last_active: int | None = None
        self.scroll = -1
        self.grid = grid

        self.mouse_handlers: dict[int, Callable[[], NotImplementedOrNone] | None] = {}
        self.tab_widths: list[int] = []
        # Caches
        self.render_tab = lru_cache(self._render_tab)
        self._content_cache: SimpleCache = SimpleCache(maxsize=50)

    @property
    def tabs(self) -> list[TabBarTab]:
        """Return the tab-bar's tabs."""
        if callable(self._tabs):
            return list(self._tabs())
        else:
            return list(self._tabs)

    @tabs.setter
    def tabs(self, tabs: Sequence[TabBarTab]) -> None:
        """Set the tab bar's current tabs."""
        self._tabs = tabs

    @property
    def active(self) -> int:
        """Return the index of the active tab."""
        current_active = self._active() if callable(self._active) else self._active

        # Check if active tab has changed
        if self._last_active != current_active:
            # Handle tab switching
            if self._last_active is not None and 0 <= self._last_active < len(
                self.tabs
            ):
                old_tab = self.tabs[self._last_active]
                if callable(on_deactivate := old_tab.on_deactivate):
                    on_deactivate()

            # Call on_activate for new tab
            if current_active is not None and 0 <= current_active < len(self.tabs):
                new_tab = self.tabs[current_active]
                if callable(on_activate := new_tab.on_activate):
                    on_activate()

            # Ensure active tab is visible
            self.scroll_to(current_active)

            # Update last known active value
            self._last_active = current_active

        return current_active

    @active.setter
    def active(self, active: int | Callable[[], int]) -> None:
        """Set the currently active tab."""
        # Store new active value
        self._active = active

        # If it's a direct value (not callable), handle tab switching immediately
        if not callable(active):
            # Force property getter to handle the change
            _ = self.active

    def preferred_width(self, max_available_width: int) -> int | None:
        """Return the preferred width of the tab-bar control, the maximum available."""
        return max_available_width

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Return the preferred height of the tab-bar control (2 rows)."""
        return 2

    def is_focusable(self) -> bool:
        """Tell whether this user control is focusable."""
        return False

    def create_content(self, width: int, height: int) -> UIContent:
        """Generate the formatted text fragments which make the controls output."""
        self.available_width = width

        def get_content() -> tuple[
            UIContent, dict[int, Callable[[], NotImplementedOrNone] | None]
        ]:
            *fragment_lines, mouse_handlers = self.render(width)

            return UIContent(
                get_line=lambda i: fragment_lines[i],
                line_count=len(fragment_lines),
                show_cursor=False,
            ), mouse_handlers

        key = (hash(tuple(self.tabs)), width, self.active, self.scroll)
        ui_content, self.mouse_handlers = self._content_cache.get(key, get_content)
        return ui_content

    def scroll_to(self, active: int) -> None:
        """Adjust scroll position to ensure the active tab is visible."""
        # Calculate position of active tab
        pos = self.spacing  # Initial spacing
        for i in range(len(self.tabs)):
            if i == active:
                # Found active tab - check if it's visible
                tab_width = self.tab_widths[i] if i < len(self.tab_widths) else 0
                # Scroll left if tab start is before visible area
                if pos < self.scroll:
                    self.scroll = pos - self.spacing - 1
                # Scroll right if tab end is after visible area
                elif pos + tab_width > self.scroll + self.available_width:
                    self.scroll = pos + tab_width - self.available_width + 2
                break

            # Add tab width and spacing
            pos += (
                self.tab_widths[i] if i < len(self.tab_widths) else 0
            ) + self.spacing

    def _render_tab(
        self,
        title: tuple[OneStyleAndTextTuple, ...],
        on_activate: Callable[[], NotImplementedOrNone],
        on_deactivate: Callable[[], NotImplementedOrNone] | None,
        on_close: Callable[[], NotImplementedOrNone] | None,
        closeable: bool,
        active: bool,
        max_title_width: int,
        grid: GridStyle,
    ) -> tuple[
        StyleAndTextTuples,
        StyleAndTextTuples,
        list[Callable[[], NotImplementedOrNone] | None],
    ]:
        """Render the tab as formatted text.

        Args:
            title: The formatted text fragments making up the tab's title
            on_activate: Callback function to run when the tab is activated
            on_deactivate: Optional callback function to run when the tab is deactivated
            on_close: Optional callback function to run when the tab is closed
            closeable: Whether the tab can be closed
            active: Whether this tab is currently active
            max_title_width: Maximum width to display the tab title
            grid: The grid style to use for drawing borders

        Returns:
            Tuple containing:
            - Top line formatted text fragments
            - Bottom line formatted text fragments
            - List of mouse handler callbacks for each character position
        """
        title_ft = truncate(list(title), max_title_width)
        title_width = fragment_list_width(title_ft)
        style = "class:active" if active else "class:inactive"

        top_line: StyleAndTextTuples = explode_text_fragments([])
        tab_line: StyleAndTextTuples = explode_text_fragments([])
        mouse_handlers: list[Callable[[], NotImplementedOrNone] | None] = []

        # Add top edge over title
        top_line.append(
            (f"{style} class:tab,border,top", grid.TOP_MID * (title_width + 2))
        )

        # Left edge
        tab_line.append((f"{style} class:tab,border,left", grid.MID_LEFT))
        mouse_handlers.append(on_activate)

        # Title
        tab_line.extend(
            [
                (f"{style} class:tab,title {frag_style}", text)
                for frag_style, text, *_ in title_ft
            ]
        )
        for _ in range(title_width):
            mouse_handlers.append(on_activate)

        # Close button
        if closeable:
            top_line.append((f"{style} class:tab,border,top", grid.TOP_MID * 2))
            mouse_handlers.append(on_activate)
            tab_line.extend(
                [
                    (f"{style} class:tab", " "),
                    (f"{style} class:tab,close", self.char_close),
                ]
            )
            mouse_handlers.append(on_close)

        # Right edge
        tab_line.append((f"{style} class:tab,border,right", grid.MID_RIGHT))
        mouse_handlers.append(on_activate)

        return (top_line, tab_line, mouse_handlers)

    def render(
        self, width: int
    ) -> tuple[
        StyleAndTextTuples,
        StyleAndTextTuples,
        dict[int, Callable[[], NotImplementedOrNone] | None],
    ]:
        """Render the tab-bar as lines of formatted text."""
        top_line: StyleAndTextTuples = []
        tab_line: StyleAndTextTuples = []
        mouse_handlers: dict[int, Callable[[], NotImplementedOrNone] | None] = {}
        pos = 0
        full = 0

        renderings = [
            self.render_tab(
                title=tuple(to_formatted_text(tab.title)),
                max_title_width=self.max_title_width,
                on_activate=tab.on_activate,
                on_deactivate=tab.on_deactivate,
                on_close=tab.on_close,
                closeable=tab.closeable,
                active=(self.active == j),
                grid=self.grid,
            )
            for j, tab in enumerate(self.tabs)
        ]
        self.tab_widths = [len(x[0]) for x in renderings]

        # Do initial scroll if first render
        if self.scroll == -1:
            self.scroll_to(self._active() if callable(self._active) else self._active)

        # Apply scroll limits
        self.scroll = max(
            0,
            min(
                self.scroll,
                self.spacing * (len(self.tabs) + 1)
                + sum(len(x[1]) for x in renderings)
                - width,
            ),
        )
        scroll = self.scroll

        # Initial spacing
        for _ in range(self.spacing):
            if full >= scroll:
                top_line += [("", " ")]
                tab_line += [("class:border,bottom", self.grid.TOP_MID)]
                pos += 1
            full += 1

        for rendering in renderings:
            # Add the rendered tab content
            for tab_top, tab_bottom, handler in zip(*rendering):
                if full >= scroll:
                    top_line.append(tab_top)
                    tab_line.append(tab_bottom)
                    mouse_handlers[pos] = handler
                    pos += 1
                full += 1
                if pos == width:
                    break

            if pos == width:
                break

            # Inter-tab spacing
            if rendering is not renderings[-1]:
                if full >= scroll:
                    for _ in range(self.spacing):
                        top_line += [("", " ")]
                        tab_line += [("class:border,bottom", self.grid.TOP_MID)]
                        if pos == width:
                            break
                        pos += 1
                full += 1

            if pos == width:
                break

        # Add scroll indicators
        if scroll > 0:
            top_line[0] = ("", " ")
            tab_line[0] = ("class:overflow", self.char_scroll_left)
        if pos >= width:
            top_line[-1] = ("", " ")
            tab_line[-1] = ("class:overflow", self.char_scroll_right)
        else:
            # Otherwise add border to fill width
            tab_line += [
                (
                    "class:border,bottom",
                    self.grid.TOP_MID * (width - pos + 1),
                )
            ]

        return top_line, tab_line, mouse_handlers

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events."""
        row = mouse_event.position.y
        col = mouse_event.position.x

        if row == 1 and mouse_event.event_type == MouseEventType.MOUSE_UP:
            if mouse_event.button == MouseButton.LEFT and callable(
                handler := self.mouse_handlers.get(col)
            ):
                # Activate the tab
                handler()
                return None
            elif mouse_event.button == MouseButton.MIDDLE:
                if callable(handler := self.mouse_handlers.get(col)):
                    # Activate tab
                    handler()
                    # Close the now active tab
                    tabs = self.tabs
                    tab = tabs[self.active]
                    if tab.closeable and callable(tab.on_close):
                        tab.on_close()
                return None

        tabs = self.tabs
        if mouse_event.event_type == MouseEventType.SCROLL_UP:
            index = max(self.active - 1, 0)
            if index != self.active:
                if callable(activate := tabs[index].on_activate):
                    activate()
                return None
        elif mouse_event.event_type == MouseEventType.SCROLL_DOWN:
            index = min(self.active + 1, len(tabs) - 1)
            if index != self.active:
                if callable(activate := tabs[index].on_activate):
                    activate()
                return None
        return NotImplemented


class StackedSplit(metaclass=ABCMeta):
    """Base class for containers with selectable children."""

    def __init__(
        self,
        children: Sequence[AnyContainer],
        titles: Sequence[AnyFormattedText],
        active: int = 0,
        style: str | Callable[[], str] = "class:tab-split",
        on_change: Callable[[StackedSplit], None] | None = None,
        width: AnyDimension = None,
        height: AnyDimension = None,
    ) -> None:
        """Create a new tabbed container instance.

        Args:
            children: A list of child container or a callable which returns such
            titles: A list of tab titles or a callable which returns such
            active: The index of the active tab
            style: A style to apply to the tabbed container
            on_change: Callback to run when the selected tab changes
            width: The width of the split container
            height: The height of the split container
        """
        self._children = list(children)
        self._titles = list(titles)
        self._active: int | None = active
        self.style = style
        self.on_change = Event(self, on_change)
        self.width = width
        self.height = height

        self.container = self.load_container()

    @abstractmethod
    def load_container(self) -> AnyContainer:
        """Abstract method for loading the widget's container."""
        ...

    def add_style(self, style: str) -> str:
        """Add a style to the widget's base style."""
        base_style = self.style() if callable(self.style) else self.style
        return f"{base_style} {style}"

    @property
    def active(self) -> int | None:
        """Return the index of the active child container."""
        return self._active

    @active.setter
    def active(self, value: int | None) -> None:
        """Set the active child container.

        Args:
            value: The index of the tab to make active
        """
        if value is not None:
            value = max(0, min(value, len(self.children) - 1))
        if value != self._active:
            self._active = value
            self.refresh()
            self.on_change.fire()
            if value is not None:
                try:
                    get_app().layout.focus(self.children[value])
                except ValueError:
                    pass

    @property
    def children(self) -> list[AnyContainer]:
        """Return a list of the widget's child containers."""
        return self._children

    @children.setter
    def children(self, value: Sequence[AnyContainer]) -> None:
        """Set the widget's child containers."""
        self._children = list(value)
        self.refresh()

    def active_child(self) -> AnyContainer:
        """Return the currently active child container."""
        return self.children[self.active or 0]

    @property
    def titles(self) -> list[AnyFormattedText]:
        """Return the titles of the child containers."""
        return self._titles

    @titles.setter
    def titles(self, value: Sequence[AnyFormattedText]) -> None:
        """Set the titles of the child containers."""
        self._titles = list(value)
        self.refresh()

    def refresh(self) -> None:
        """Reload the widget's container when its children or their titles change."""

    def __pt_container__(self) -> AnyContainer:
        """Return the widget's container."""
        return self.container


class TabbedSplit(StackedSplit):
    """A container which switches between children using tabs."""

    def __init__(
        self,
        children: Sequence[AnyContainer],
        titles: Sequence[AnyFormattedText],
        active: int = 0,
        style: str | Callable[[], str] = "class:tab-split",
        on_change: Callable[[StackedSplit], None] | None = None,
        width: AnyDimension = None,
        height: AnyDimension = None,
        border: GridStyle = OutsetGrid,
        show_borders: DiBool | None = None,
    ) -> None:
        """Initialize a new tabbed container."""
        self.border = border
        self.show_borders = show_borders or DiBool(False, True, True, True)

        kb = KeyBindings()

        @kb.add("left")
        def _prev(event: KeyPressEvent) -> None:
            """Previous tab."""
            self.active = (self.active or 0) - 1

        @kb.add("right")
        def _next(event: KeyPressEvent) -> None:
            """Next tab."""
            self.active = (self.active or 0) + 1

        self.key_bindings = kb

        super().__init__(
            children=children,
            titles=titles,
            active=active,
            style=style,
            on_change=on_change,
            width=width,
            height=height,
        )

    def load_container(self) -> AnyContainer:
        """Create the tabbed widget's container.

        Consists of a tab-bar control above a dynamic container which shows the active
        child container.

        Returns:
            The widget's container
        """
        self.control = TabBarControl(self.load_tabs(), active=self.active or 0)
        return HSplit(
            [
                Window(
                    self.control,
                    style=partial(self.add_style, "class:tab-bar"),
                    height=2,
                ),
                Border(
                    Box(
                        DynamicContainer(self.active_child),
                        padding=0,
                        padding_top=1,
                        padding_bottom=1,
                        style="class:tabbed-split,page",
                    ),
                    border=self.border,
                    show_borders=self.show_borders,
                ),
            ],
            style="class:tabbed-split",
            width=self.width,
            height=self.height,
            key_bindings=self.key_bindings,
        )

    def refresh(self) -> None:
        """Refresh the widget - set the tab-bar's tabs and active tab index."""
        self.control.tabs = self.load_tabs()
        if self.active is not None:
            self.control.active = self.active

    def load_tabs(self) -> list[TabBarTab]:
        """Return a list of tabs for the current children."""
        return [
            TabBarTab(
                title=title,
                on_activate=partial(setattr, self, "active", i),
            )
            for i, title in enumerate(self.titles)
        ]


class AccordionSplit(StackedSplit):
    """A container which switches between children using expandable sections."""

    def load_container(self) -> AnyContainer:
        """Create the accordiion widget's container."""
        self.draw_container()
        return DynamicContainer(lambda: self._container)

    def draw_container(self) -> None:
        """Render the accordion in it's current state."""
        self._container = HSplit(
            [
                Border(
                    HSplit(
                        [
                            Window(
                                FormattedTextControl(
                                    partial(self.title_text, index, title),
                                    focusable=True,
                                    show_cursor=False,
                                )
                            ),
                            ConditionalContainer(
                                Box(child, padding_left=0),
                                filter=Condition(
                                    partial(lambda i: self.active == i, index)
                                ),
                            ),
                        ]
                    ),
                    style=partial(self.add_style, "class:border"),
                )
                for index, (title, child) in enumerate(zip(self.titles, self.children))
            ],
            style="class:accordion",
        )

    def title_text(self, index: int, title: AnyFormattedText) -> StyleAndTextTuples:
        """Generate the title for each child container."""
        return [
            ("", " "),
            (
                "bold" + (" class:selection" if self.active == index else ""),
                "▶" if self.active == index else "▼",
                cast(
                    "Callable[[MouseEvent], None]", partial(self.mouse_handler, index)
                ),
            ),
            ("", " "),
            *[
                (
                    f"bold {style}",
                    text,
                    cast(
                        "Callable[[MouseEvent], None]",
                        partial(self.mouse_handler, index),
                    ),
                )
                for style, text, *_ in to_formatted_text(title)
            ],
        ]

    def mouse_handler(
        self, index: int, mouse_event: MouseEvent
    ) -> NotImplementedOrNone:
        """Handle mouse events."""
        # if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
        #    get_app().layout.focus()
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            self.toggle(index)
        else:
            return NotImplemented
        return None

    def toggle(
        self,
        index: int,
    ) -> None:
        """Toggle the visibility of a child container."""
        self.active = index if self.active != index else None

    def refresh(self) -> None:
        """Re-draw the container when the list of child containers changes."""
        self.draw_container()



"""Contains a completion menu for toolbars."""

from __future__ import annotations

import logging
from math import ceil
from typing import TYPE_CHECKING

from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import Condition, has_completions, is_done
from prompt_toolkit.layout.containers import ConditionalContainer, Window
from prompt_toolkit.layout.controls import GetLinePrefixCallable, UIContent, UIControl
from prompt_toolkit.layout.dimension import Dimension
from prompt_toolkit.utils import get_cwidth

from euporie.core.app.current import get_app
from euporie.core.filters import has_toolbar
from euporie.core.ft.utils import apply_style, pad, truncate
from euporie.core.layout.containers import HSplit

if TYPE_CHECKING:
    from prompt_toolkit.buffer import CompletionState
    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import (
        KeyBindingsBase,
        NotImplementedOrNone,
    )
    from prompt_toolkit.mouse_events import MouseEvent

log = logging.getLogger(__name__)


class ToolbarCompletionMenuControl(UIControl):
    """A completion menu for toolbars."""

    def __init__(self, min_item_width: int = 5, max_item_width: int = 30) -> None:
        """Define minimum and maximum item widhth."""
        self.max_item_width = max_item_width
        self.min_item_width = min_item_width

    def preferred_width(self, max_available_width: int) -> int | None:
        """Fill available width."""
        return max_available_width

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Calculate how many rows to use, filling the width first then overflowing."""
        complete_state = get_app().current_buffer.complete_state
        if complete_state is None:
            return 0

        col_width = self._get_col_width(complete_state, width, max_available_height)
        height = min(
            ceil(col_width * len(complete_state.completions) / width),
            max_available_height,
        )
        return height

    def _get_col_width(
        self, complete_state: CompletionState, width: int, height: int
    ) -> int:
        """Calculate the optimal width for the items in the menu."""
        completions = complete_state.completions
        item_width = max(
            min(
                max(get_cwidth(c.display_text) + 3 for c in completions),
                self.max_item_width,
            ),
            self.min_item_width,
        )
        col_count = width // item_width
        # With an overflow reduce column width to show more of the truncated column
        if len(completions) > col_count * height:
            col_width = min((width - 6) // col_count, item_width)
        # With an exact width expand columns to fill the space
        elif len(completions) == col_count * height:
            col_width = max(width // col_count, item_width)
        # Otherwise use the calculated item width
        else:
            col_width = item_width
        return col_width

    def create_content(self, width: int, height: int) -> UIContent:
        """Create a UIContent object for this control."""
        complete_state = get_app().current_buffer.complete_state
        if complete_state is None:
            return UIContent()

        completions = complete_state.completions
        index = complete_state.complete_index  # Can be None!

        # Calculate width of completions menu.
        col_width = self._get_col_width(complete_state, width, height)
        # Calculate offset to ensure active completion is visible
        cur_col = (index or 0) // height
        visible_cols = width // col_width
        offset = max(0, cur_col - visible_cols + 1) * height

        # Pad and style visible items
        items: list[StyleAndTextTuples] = []
        item: StyleAndTextTuples
        for i in range(offset, offset + ((visible_cols + 1) * height)):
            if i < len(completions):
                item = completions[i].display
                item = truncate(item, col_width - 3)
                item = pad(item, width=col_width - 3)
                item = [("", " "), *item, ("", " ")]
                item = apply_style(item, "class:completion")
                if i == index:
                    item = apply_style(item, "class:current")
                item = [*item, ("", " ")]
            else:
                item = [("", " " * col_width)]
            items.append(item)

        # Construct rows
        overflow_left = offset > height
        overflow_right = (len(completions) - offset) - (visible_cols * height) > 0
        lines: list[StyleAndTextTuples] = (
            [
                [("class:overflow", "◀" if i == height // 2 else " ")]
                for i in range(height)
            ]
            if overflow_left
            else [[] for _ in range(height)]
        )
        for i, item in enumerate(items):
            row = i % height
            col = i // height
            if col == visible_cols:
                item = [
                    *truncate(
                        item,
                        width - overflow_left - overflow_right - col_width * col,
                        placeholder="",
                    ),
                    ("class:overflow", "▶" if row == height // 2 else " "),
                ]
            lines[row].extend(item)

        def get_line(i: int) -> StyleAndTextTuples:
            return lines[i]

        return UIContent(
            get_line=get_line,
            cursor_position=Point(x=0, y=0),  # y=index or 0),
            line_count=len(lines),
        )

    def mouse_handler(self, mouse_event: MouseEvent) -> NotImplementedOrNone:
        """Handle mouse events.

        When `NotImplemented` is returned, it means that the given event is not
        handled by the `UIControl` itself. The `Window` or key bindings can
        decide to handle this event as scrolling or changing focus.

        :param mouse_event: `MouseEvent` instance.
        """
        return NotImplemented

    def move_cursor_down(self) -> None:
        """Request to move the cursor down.

        This happens when scrolling down and the cursor is completely at the top.
        """

    def move_cursor_up(self) -> None:
        """Request to move the cursor up."""

    def get_key_bindings(self) -> KeyBindingsBase | None:
        """Key bindings that are specific for this user control.

        Return a :class:`.KeyBindings` object if some key bindings are
        specified, or `None` otherwise.
        """


class SelectedCompletionMetaControl(UIControl):
    """Control that shows the meta information of the selected completion."""

    def preferred_width(self, max_available_width: int) -> int | None:
        """Report the width of the active meta text."""
        if (
            (state := get_app().current_buffer.complete_state)
            and (current_completion := state.current_completion)
            and (text := current_completion.display_meta_text)
        ):
            return get_cwidth(text) + 2
        return 0

    def preferred_height(
        self,
        width: int,
        max_available_height: int,
        wrap_lines: bool,
        get_line_prefix: GetLinePrefixCallable | None,
    ) -> int | None:
        """Maintain a single line."""
        return 1

    def create_content(self, width: int, height: int) -> UIContent:
        """Format the current completion meta text."""
        ft: StyleAndTextTuples = []
        state = get_app().current_buffer.complete_state
        if (
            (state := get_app().current_buffer.complete_state)
            and (current_completion := state.current_completion)
            and (meta := current_completion.display_meta)
        ):
            ft = apply_style([("", " "), *meta, ("", " ")], style="class:meta")

        def get_line(i: int) -> StyleAndTextTuples:
            return ft

        return UIContent(get_line=get_line, line_count=1 if ft else 0)


class ToolbarCompletionsMenu(ConditionalContainer):
    """A completion menu widget for toolbars."""

    def __init__(self) -> None:
        """Create a pre-populated conditional container."""
        super().__init__(
            content=HSplit(
                [
                    ConditionalContainer(
                        Window(
                            content=SelectedCompletionMetaControl(),
                            height=1,
                            dont_extend_width=True,
                        ),
                        filter=Condition(
                            lambda: bool(
                                (
                                    complete_state
                                    := get_app().current_buffer.complete_state
                                )
                                and (completion := complete_state.current_completion)
                                and completion.display_meta
                            )
                        ),
                    ),
                    Window(
                        content=ToolbarCompletionMenuControl(),
                        height=Dimension(min=1, max=8),
                        dont_extend_height=True,
                        dont_extend_width=False,
                    ),
                ],
                style="class:toolbar,menu",
            ),
            filter=has_toolbar & has_completions & ~is_done,
        )



"""Define a tree-view widget."""

from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.mouse_events import MouseButton, MouseEvent, MouseEventType

if TYPE_CHECKING:
    from typing import Any

    from prompt_toolkit.formatted_text import StyleAndTextTuples
    from prompt_toolkit.key_binding.key_bindings import NotImplementedOrNone


class JsonView:
    """A JSON-view container."""

    def __init__(
        self, data: Any, title: str | None = None, expanded: bool = False
    ) -> None:
        """Create a new instance."""
        self.data = data
        self.title = "root" if title is None else title
        self.start_expanded = expanded
        self._toggled_paths: set[tuple[str, ...]] = set()
        self.container = Window(
            FormattedTextControl(self._get_formatted_text, focusable=True),
            style="class:tree",
        )

    def _get_value_style(self, value: Any) -> str:
        """Return the style for a given value type."""
        return {
            str: "class:pygments.literal.string",
            int: "class:pygments.literal.number",
            float: "class:pygments.literal.number",
            bool: "class:pygments.keyword.constant",
        }.get(type(value), "")

    def _format_value(self, value: Any) -> tuple[str, str]:
        """Return the formatted value and its style."""
        if isinstance(value, (list, dict)):
            length = len(value)
            if isinstance(value, list):
                return f" [] {length} items", "class:pygments.comment"
            return f" {{}} {length} items", "class:pygments.comment"
        return repr(value), self._get_value_style(value)

    def _get_formatted_text(self) -> StyleAndTextTuples:
        """Generate the complete tree view as formatted text."""
        result: StyleAndTextTuples = []
        toggled_paths = self._toggled_paths
        start_expanded = self.start_expanded
        toggle = self._toggle
        format_value = self._format_value

        def format_node(
            data: Any, path: tuple[str, ...], indent: int, key: str
        ) -> None:
            is_expanded = (path in toggled_paths) ^ start_expanded
            has_children = isinstance(data, (dict, list))
            mouse_handler = partial(toggle, path=path)

            value, style = format_value(data)
            row: StyleAndTextTuples = [
                # Add indentation
                ("", "  " * indent),
                # Add toggle symbol
                (
                    ("class:pygments.operator", "▼" if is_expanded else "▶")
                    if has_children
                    else ("", " ")
                ),
                ("", " "),
                ("class:pygments.keyword", str(key)),
                ("class:pygments.punctuation", ": "),
                (style, value),
                ("", "\n"),
            ]

            # Apply mouse_handler to rows with children
            if has_children:
                row = [(style, text, mouse_handler) for (style, text, *_) in row]

            result.extend(row)

            if is_expanded and has_children:
                if isinstance(data, list):
                    data = {str(i): v for i, v in enumerate(data)}

                for k, v in data.items():
                    new_path = (*path, str(k)) if path else (str(k),)
                    format_node(v, new_path, indent + 1, k)

        format_node(self.data, (), 0, self.title)
        return result

    def _toggle(
        self, mouse_event: MouseEvent, path: tuple[str, ...]
    ) -> NotImplementedOrNone:
        """Toggle the expansion state of a node."""
        if (
            mouse_event.button == MouseButton.LEFT
            and mouse_event.event_type == MouseEventType.MOUSE_UP
        ):
            if path in self._toggled_paths:
                self._toggled_paths.remove(path)
            else:
                self._toggled_paths.add(path)
            return None
        return NotImplemented

    def __pt_container__(self) -> Window:
        """Return the tree-view container's content."""
        return self.container



"""Contains key handlers for terminal queries."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from euporie.core.commands import add_cmd
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindingsBase, KeyPressEvent

    from euporie.core.config import Config

log = logging.getLogger(__name__)

_COLOR_NAMES: dict[str, str] = {
    "10": "fg",
    "11": "bg",
    "4;0": "ansiblack",
    "4;1": "ansired",
    "4;2": "ansigreen",
    "4;3": "ansiyellow",
    "4;4": "ansiblue",
    "4;5": "ansipurple",
    "4;6": "ansicyan",
    "4;7": "ansiwhite",
    "4;8": "ansirbightblack",
    "4;9": "ansirbightred",
    "4;10": "ansirbightgreen",
    "4;11": "ansirbightyellow",
    "4;12": "ansirbightblue",
    "4;13": "ansirbightpurple",
    "4;14": "ansirbightcyan",
    "4;15": "ansirbightwhite",
}


def get_match(event: KeyPressEvent) -> dict[str, str] | None:
    """Get pattern matches from a key press event."""
    if (
        (parser := getattr(event.app.input, "vt100_parser", None))
        and (patterns := getattr(parser, "patterns", None))
        and (pattern := patterns.get(event.key_sequence[-1].key))
        and (match := pattern.match(event.data))
        and (values := match.groupdict())
    ):
        return values
    return None


@add_cmd(hidden=True, is_global=True)
def _set_terminal_color(event: KeyPressEvent) -> object:
    """Run when the terminal receives a terminal color query response.

    Args:
        event: The key press event received when the termina sends a response

    Returns:
        :py:obj:`NotImplemented`, so the application is not invalidated when a
        response from the terminal is received

    """
    from euporie.core.app.app import BaseApp

    if isinstance(app := event.app, BaseApp) and (colors := get_match(event)):
        c = colors["c"]
        r, g, b = colors.get("r", "00"), colors.get("g", "00"), colors.get("b", "00")
        app.term_colors[_COLOR_NAMES.get(c, c)] = f"#{r[:2]}{g[:2]}{b[:2]}"
        app.update_style()
    return NotImplemented


@add_cmd(hidden=True, is_global=True)
def _set_terminal_pixel_size(event: KeyPressEvent) -> object:
    """Run when the terminal receives a pixel dimension query response."""
    from euporie.core.app.app import BaseApp

    if (
        isinstance(app := event.app, BaseApp)
        and (values := get_match(event))
        and (x := values.get("x"))
        and (y := values.get("y"))
    ):
        app.term_size_px = int(x), int(y)
    return NotImplemented


@add_cmd(hidden=True, is_global=True)
def _set_terminal_graphics_sixel(event: KeyPressEvent) -> object:
    """Run when the terminal receives a sixel graphics support query response."""
    from euporie.core.app.app import BaseApp

    if (
        isinstance(app := event.app, BaseApp)
        and (values := get_match(event))
        and values.get("sixel")
    ):
        app.term_graphics_sixel = True
    return NotImplemented


@add_cmd(hidden=True, is_global=True)
def _set_terminal_graphics_iterm(event: KeyPressEvent) -> object:
    """Run when the terminal receives a iterm graphics support query response."""
    from euporie.core.app.app import BaseApp

    if (
        isinstance(app := event.app, BaseApp)
        and (values := get_match(event))
        and (term := values.get("term"))
        and term.startswith(("WezTerm", "Konsole", "mlterm"))
    ):
        app.term_graphics_iterm = True
    return NotImplemented


@add_cmd(hidden=True, is_global=True)
def _set_terminal_graphics_kitty(event: KeyPressEvent) -> object:
    """Run when the terminal receives a kitty graphics support query response."""
    from euporie.core.app.app import BaseApp

    if (
        isinstance(app := event.app, BaseApp)
        and (values := get_match(event))
        and values.get("status") == "OK"
    ):
        app.term_graphics_kitty = True
    return NotImplemented


@add_cmd(hidden=True, is_global=True)
def _set_terminal_sgr_pixel(event: KeyPressEvent) -> object:
    """Run when the terminal receives a SGR-pixel mode support query response."""
    from euporie.core.app.app import BaseApp

    if (
        isinstance(app := event.app, BaseApp)
        and (values := get_match(event))
        and (values.get("Pm") in {"1", "3"})
    ):
        app.term_sgr_pixel = True
    return NotImplemented


@add_cmd(hidden=True, is_global=True)
def _set_terminal_clipboard_data(event: KeyPressEvent) -> object:
    """Run when the terminal receives a clipboard data query response."""
    from base64 import b64decode

    from euporie.core.app.app import BaseApp
    from euporie.core.clipboard import Osc52Clipboard

    if (
        isinstance(app := event.app, BaseApp)
        and isinstance(clipboard := app.clipboard, Osc52Clipboard)
        and (values := get_match(event))
    ):
        value = values.get("data", "")
        text = b64decode(value).decode()
        log.warning(repr(text))
        clipboard.sync(text)
    return NotImplemented


class TerminalQueries:
    """Key bindings for terminal query responses."""


register_bindings(
    {
        "euporie.core.io.TerminalInfo": {
            "set-terminal-color": "<colors-response>",
            "set-terminal-pixel-size": "<pixel-size-response>",
            "set-terminal-graphics-kitty": "<kitty-graphics-status-response>",
            "set-terminal-graphics-sixel": "<sixel-graphics-status-response>",
            "set-terminal-graphics-iterm": "<iterm-graphics-status-response>",
            "set-terminal-sgr-pixel": "<sgr-pixel-status-response>",
            "set-terminal-clipboard-data": "<clipboard-data-response>",
        }
    }
)


def load_terminal_bindings(config: Config | None = None) -> KeyBindingsBase:
    """Load editor key-bindings in the style of the ``micro`` text editor."""
    return load_registered_bindings(
        "euporie.core.key_binding.bindings.terminal.TerminalQueries", config=config
    )



"""Contain the main Application class which runs euporie.core."""

from __future__ import annotations

import asyncio
import logging
import os
import signal
import sys
from abc import ABC, abstractmethod
from enum import Enum
from functools import partial
from pathlib import PurePath
from typing import TYPE_CHECKING, cast
from weakref import WeakSet, WeakValueDictionary

from prompt_toolkit.application.application import Application, _CombinedRegistry
from prompt_toolkit.application.current import create_app_session, set_app
from prompt_toolkit.data_structures import Point
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import Condition, buffer_has_focus, to_filter
from prompt_toolkit.input.defaults import create_input
from prompt_toolkit.key_binding.bindings.basic import (
    load_basic_bindings as load_ptk_basic_bindings,
)
from prompt_toolkit.key_binding.bindings.cpr import load_cpr_bindings
from prompt_toolkit.key_binding.bindings.emacs import (
    load_emacs_bindings,
    load_emacs_search_bindings,
    load_emacs_shift_selection_bindings,
)
from prompt_toolkit.key_binding.bindings.mouse import (
    load_mouse_bindings as load_ptk_mouse_bindings,
)
from prompt_toolkit.key_binding.bindings.vi import load_vi_search_bindings
from prompt_toolkit.key_binding.key_bindings import (
    ConditionalKeyBindings,
    merge_key_bindings,
)
from prompt_toolkit.layout.containers import Float, FloatContainer, to_container
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.output import ColorDepth
from prompt_toolkit.output.defaults import create_output
from prompt_toolkit.output.vt100 import Vt100_Output as PtkVt100_Output
from prompt_toolkit.styles import (
    BaseStyle,
    ConditionalStyleTransformation,
    DummyStyle,
    SetDefaultColorStyleTransformation,
    Style,
    SwapLightAndDarkStyleTransformation,
    merge_style_transformations,
    merge_styles,
    style_from_pygments_cls,
)
from prompt_toolkit.utils import Event

from euporie.core.app.base import ConfigurableApp
from euporie.core.app.cursor import CursorConfig
from euporie.core.clipboard import CONFIGURED_CLIPBOARDS
from euporie.core.convert.mime import get_mime
from euporie.core.filters import has_toolbar
from euporie.core.format import CliFormatter
from euporie.core.io import COLOR_DEPTHS, Vt100_Output, Vt100Parser
from euporie.core.key_binding.key_processor import KeyProcessor
from euporie.core.key_binding.micro_state import MicroState
from euporie.core.key_binding.registry import (
    load_registered_bindings,
    register_bindings,
)
from euporie.core.key_binding.vi_state import ViState
from euporie.core.layout.containers import Window
from euporie.core.log import setup_logs
from euporie.core.lsp import KNOWN_LSP_SERVERS, LspClient
from euporie.core.path import parse_path
from euporie.core.renderer import Renderer
from euporie.core.style import (
    DEFAULT_COLORS,
    HTML_STYLE,
    IPYWIDGET_STYLE,
    LOG_STYLE,
    MIME_STYLE,
    ColorPalette,
    build_style,
    get_style_by_name,
)
from euporie.core.utils import ChainedList
from euporie.core.widgets.decor import Shadow
from euporie.core.widgets.menu import CompletionsMenu

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    from pathlib import Path
    from types import FrameType
    from typing import Any, Callable, ClassVar, TypeVar

    # from prompt_toolkit.application import _AppResult
    from prompt_toolkit.contrib.ssh import PromptToolkitSSHSession
    from prompt_toolkit.filters import Filter, FilterOrBool
    from prompt_toolkit.input import Input
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.layout.layout import FocusableElement
    from prompt_toolkit.layout.screen import WritePosition
    from prompt_toolkit.output import Output

    from euporie.core.bars.command import CommandBar
    from euporie.core.bars.search import SearchBar
    from euporie.core.config import Setting
    from euporie.core.format import Formatter
    from euporie.core.tabs import TabRegistryEntry
    from euporie.core.tabs.base import Tab
    from euporie.core.widgets.dialog import Dialog
    from euporie.core.widgets.pager import Pager

    _AppResult = TypeVar("_AppResult")

log = logging.getLogger(__name__)


class ExtraEditingMode(str, Enum):
    """Additional editing modes."""

    MICRO = "MICRO"


class BaseApp(ConfigurableApp, Application, ABC):
    """All euporie apps.

    The base euporie application class.

    This subclasses the `prompt_toolkit.application.Application` class, so application
    wide methods can be easily added.
    """

    color_palette: ColorPalette
    mouse_position: Point

    _config_defaults: ClassVar[dict[str, Any]] = {"log_level_stdout": "critical"}

    def __init__(
        self,
        title: str | None = None,
        set_title: bool = True,
        leave_graphics: FilterOrBool = True,
        extend_renderer_height: FilterOrBool = False,
        extend_renderer_width: FilterOrBool = False,
        enable_page_navigation_bindings: FilterOrBool | None = True,
        **kwargs: Any,
    ) -> None:
        """Instantiate euporie specific application variables.

        After euporie specific application variables are instantiated, the application
        instance is initiated.

        Args:
            title: The title string to set in the terminal
            set_title: Whether to set the terminal title
            leave_graphics: A filter which determines if graphics should be cleared
                from the display when they are no longer active
            extend_renderer_height: Whether the renderer height should be extended
                beyond the height of the display
            extend_renderer_width: Whether the renderer width should be extended
                beyond the height of the display
            enable_page_navigation_bindings: Determines if page navigation keybindings
                should be loaded
            kwargs: The key-word arguments for the :py:class:`Application`

        """
        # Initialise the application
        super().__init__(
            **{
                "clipboard": CONFIGURED_CLIPBOARDS.get(
                    self.config.clipboard, lambda: None
                )(),
                "color_depth": COLOR_DEPTHS.get(self.config.color_depth),
                "editing_mode": self.get_edit_mode(),
                "mouse_support": True,
                "cursor": CursorConfig(),
                "enable_page_navigation_bindings": enable_page_navigation_bindings,
                **kwargs,
            }
        )
        # Use custom renderer
        self.renderer = Renderer(
            self._merged_style,
            self.output,
            full_screen=self.full_screen,
            mouse_support=self.mouse_support,
            cpr_not_supported_callback=self.cpr_not_supported_callback,
            extend_height=extend_renderer_height,
            extend_width=extend_renderer_width,
        )
        # Contains the opened tab containers
        self.tabs: list[Tab] = []
        self.on_tabs_change = Event(self)
        # Holds the optional toolbars
        self.search_bar: SearchBar | None = None
        self.command_bar: CommandBar | None = None
        # Holds the index of the current tab
        self._tab_idx = 0
        # Add state for micro key-bindings
        self.micro_state = MicroState()
        # Default terminal info values
        self.term_colors = dict(DEFAULT_COLORS)
        self.term_graphics_sixel = False
        self.term_graphics_iterm = False
        self.term_graphics_kitty = False
        self.term_sgr_pixel = False
        self._term_size_px: tuple[int, int]
        # Floats at the app level
        self.leave_graphics = to_filter(leave_graphics)
        self.graphics: WeakSet[Float] = WeakSet()
        self.dialogs: dict[str, Dialog] = {}
        self.menus: dict[str, Float] = {}
        self.floats = ChainedList(
            self.graphics,
            self.dialogs.values(),
            self.menus.values(),
        )
        # Continue loading when the application has been launched
        # and an event loop has been created
        self.pre_run_callables = [self.pre_run]
        self.post_load_callables: list[Callable[[], None]] = []
        # Set default vi input mode to navigate
        self.vi_state = ViState()
        # Set a long timeout for mappings (e.g. dd)
        self.timeoutlen = 1.0
        # Set a short timeout for flushing input
        self.ttimeoutlen = 0.0
        # Use a custom key-processor which does not wait after escape keys
        self.key_processor = KeyProcessor(_CombinedRegistry(self))
        # List of key-bindings groups to load
        self.bindings_to_load = [
            "euporie.core.app.app:BaseApp",
            "euporie.core.io.TerminalInfo",
        ]

        if enable_page_navigation_bindings:
            from euporie.core.key_binding.bindings.page_navigation import (
                load_page_navigation_bindings,
            )

            self._page_navigation_bindings = load_page_navigation_bindings(self.config)
        # Allow hiding element when manually redrawing app
        self._redrawing = False
        self.redrawing = Condition(lambda: self._redrawing)
        # Add an optional pager
        self.pager: Pager | None = None
        # Stores the initially focused element
        self.focused_element: FocusableElement | None = None
        # Set the terminal title
        self.set_title = to_filter(set_title)
        self.title = title or self.__class__.__name__
        # Register config hooks
        self.config.events.edit_mode += self.update_edit_mode
        self.config.events.syntax_theme += self.update_style
        self.config.events.color_scheme += self.update_style
        self.config.events.log_level += lambda x: setup_logs(self.config)
        self.config.events.log_file += lambda x: setup_logs(self.config)
        self.config.events.log_config += lambda x: setup_logs(self.config)
        self.config.events.color_depth += lambda x: setattr(
            self, "_color_depth", COLOR_DEPTHS[self.config.color_depth]
        )
        self.config.events.clipboard += lambda x: setattr(
            self, "clipboard", CONFIGURED_CLIPBOARDS[self.config.clipboard]
        )
        # Set up the color palette
        self.color_palette = ColorPalette()
        self.color_palette.add_color("fg", "#ffffff", "default")
        self.color_palette.add_color("bg", "#000000", "default")
        # Set up a write position to limit mouse events to a particular region
        self.mouse_limits: WritePosition | None = None
        self.mouse_position = Point(0, 0)

        # Store LSP client instances
        self.lsp_clients: WeakValueDictionary[str, LspClient] = WeakValueDictionary()

        # Build list of configured external formatters
        self.formatters: list[Formatter] = [
            CliFormatter(**info) for info in self.config.formatters
        ]

    @property
    def term_size_px(self) -> tuple[int, int]:
        """The dimensions of the terminal in pixels."""
        try:
            return self._term_size_px
        except AttributeError:
            from euporie.core.io import _tiocgwinsz

            _rows, _cols, px, py = _tiocgwinsz()
            self._term_size_px = (px, py)
        return self._term_size_px

    @term_size_px.setter
    def term_size_px(self, value: tuple[int, int]) -> None:
        self._term_size_px = value

    @property
    def cell_size_px(self) -> tuple[int, int]:
        """Get the pixel size of a single terminal cell."""
        px, py = self.term_size_px
        rows, cols = self.output.get_size()
        # If we can't get the pixel size, just guess wildly
        return px // cols or 10, py // rows or 20

    @property
    def title(self) -> str:
        """The application's title."""
        return self._title

    @title.setter
    def title(self, value: str) -> None:
        """Set the terminal title."""
        self._title = value
        if self.set_title():
            self.output.set_title(value)

    def pause_rendering(self) -> None:
        """Block rendering, but allows input to be processed.

        The first line prevents the display being drawn, and the second line means
        the key processor continues to process keys. We need this as we need to
        wait for the results of terminal queries which come in as key events.

        This is used to prevent flicker when we update the styles based on terminal
        feedback.
        """
        self._is_running = False
        self.renderer._waiting_for_cpr_futures.append(asyncio.Future())

    def resume_rendering(self) -> None:
        """Reume rendering the app."""
        self._is_running = True
        if futures := self.renderer._waiting_for_cpr_futures:
            futures.pop()

    def pre_run(self, app: Application | None = None) -> None:
        """Call during the 'pre-run' stage of application loading."""
        # Set the application's style
        self.update_style()
        # Load completions menu.
        self.menus["completions"] = Float(
            content=Shadow(CompletionsMenu(extra_filter=~has_toolbar)),
            xcursor=True,
            ycursor=True,
        )
        # Load the layout
        # We delay this until we have terminal responses to allow terminal graphics
        # support to be detected first
        self.layout = Layout(self.load_container(), self.focused_element)
        # Open any files we need to
        self.open_files()
        # Start polling terminal style if configured
        if self.config.terminal_polling_interval and hasattr(
            self.input, "vt100_parser"
        ):
            self.create_background_task(self._poll_terminal_colors())

    async def _poll_terminal_colors(self) -> None:
        """Repeatedly query the terminal for its background and foreground colours."""
        if isinstance(output := self.output, Vt100_Output):
            while self.config.terminal_polling_interval:
                await asyncio.sleep(self.config.terminal_polling_interval)
                output.get_colors()

    async def run_async(
        self,
        pre_run: Callable[[], None] | None = None,
        set_exception_handler: bool = True,
        handle_sigint: bool = True,
        slow_callback_duration: float = 0.5,
    ) -> _AppResult:
        """Run the application."""
        with set_app(self):
            # Use a custom vt100 parser to allow querying the terminal
            if parser := getattr(self.input, "vt100_parser", None):
                setattr(  # noqa B010
                    self.input, "vt100_parser", Vt100Parser(parser.feed_key_callback)
                )

            # Load key bindings
            self.load_key_bindings()

            if isinstance(self.output, Vt100_Output):
                # Send terminal queries
                self.output.get_colors()
                self.output.get_pixel_size()
                self.output.get_kitty_graphics_status()
                self.output.get_sixel_graphics_status()
                self.output.get_iterm_graphics_status()
                self.output.get_sgr_pixel_status()
                self.output.get_csiu_status()
                self.output.flush()

                # Read responses
                kp = self.key_processor

                def read_from_input() -> None:
                    kp.feed_multiple(self.input.read_keys())

                with self.input.raw_mode(), self.input.attach(read_from_input):
                    # Give the terminal time to respond and allow the event loop to read
                    # the terminal responses from the input
                    await asyncio.sleep(0.1)
                kp.process_keys()

        return await super().run_async(
            pre_run, set_exception_handler, handle_sigint, slow_callback_duration
        )

    @classmethod
    async def interact(cls, ssh_session: PromptToolkitSSHSession) -> None:
        """Run the app asynchronously for the hub SSH server."""
        try:
            await cls().run_async()
        except EOFError:
            pass

    @classmethod
    def load_input(cls) -> Input:
        """Create the input for this application to use.

        Ensures the TUI app always tries to run in a TTY.

        Returns:
            A prompt-toolkit input instance

        """
        input_ = create_input(always_prefer_tty=True)

        if (stdin := getattr(input_, "stdin", None)) and not stdin.isatty():
            from euporie.core.io import IgnoredInput

            input_ = IgnoredInput()

        return input_

    @classmethod
    def load_output(cls) -> Output:
        """Create the output for this application to use.

        Ensures the TUI app always tries to run in a TTY.

        Returns:
            A prompt-toolkit output instance

        """
        output = create_output(always_prefer_tty=True)

        if isinstance(output, PtkVt100_Output):
            output = Vt100_Output(
                stdout=output.stdout,
                get_size=output._get_size,
                term=output.term,
                default_color_depth=output.default_color_depth,
                enable_bell=output.enable_bell,
            )

        return output

    def post_load(self) -> None:
        """Allow subclasses to define additional loading steps."""
        # Call extra callables
        for cb in self.post_load_callables:
            cb()

    def load_key_bindings(self) -> None:
        """Load the application's key bindings."""
        from euporie.core.key_binding.bindings.basic import load_basic_bindings
        from euporie.core.key_binding.bindings.micro import load_micro_bindings
        from euporie.core.key_binding.bindings.mouse import load_mouse_bindings
        from euporie.core.key_binding.bindings.terminal import load_terminal_bindings
        from euporie.core.key_binding.bindings.vi import load_vi_bindings

        self._default_bindings = merge_key_bindings(
            [
                # Make sure that the above key bindings are only active if the
                # currently focused control is a `BufferControl`. For other controls, we
                # don't want these key bindings to intervene. (This would break "ptterm"
                # for instance, which handles 'Keys.Any' in the user control itself.)
                ConditionalKeyBindings(
                    merge_key_bindings(
                        [
                            # Load basic bindings.
                            load_ptk_basic_bindings(),
                            load_basic_bindings(),
                            # Load micro bindings
                            load_micro_bindings(config=self.config),
                            # Load emacs bindings.
                            load_emacs_bindings(),
                            load_emacs_search_bindings(),
                            load_emacs_shift_selection_bindings(),
                            # Load Vi bindings.
                            load_vi_bindings(),
                            load_vi_search_bindings(),
                        ]
                    ),
                    buffer_has_focus,
                ),
                # Active, even when no buffer has been focused.
                load_ptk_mouse_bindings(),
                load_cpr_bindings(),
                # Load extra mouse bindings
                load_mouse_bindings(),
                # Load terminal query response key bindings
                load_terminal_bindings(),
            ]
        )
        self.key_bindings = load_registered_bindings(
            *self.bindings_to_load, config=self.config
        )

    def _on_resize(self) -> None:
        """Query the terminal dimensions on a resize event."""
        if isinstance(output := self.output, Vt100_Output):
            output.get_pixel_size()
        super()._on_resize()

    @classmethod
    def launch(cls) -> None:
        """Launch the app."""
        from prompt_toolkit.utils import in_main_thread

        super().launch()
        # Run the application
        with create_app_session(input=cls.load_input(), output=cls.load_output()):
            # Create an instance of the app and run it
            app = cls()
            if in_main_thread():
                # Handle SIGTERM while the app is running
                original_sigterm = signal.getsignal(signal.SIGTERM)
                signal.signal(signal.SIGTERM, app.cleanup)
            # Set and run the app
            with set_app(app):
                try:
                    result = app.run()
                except (EOFError, KeyboardInterrupt):
                    result = None
                finally:
                    if in_main_thread():
                        signal.signal(signal.SIGTERM, original_sigterm)
                    # Shut down any remaining LSP clients at exit
                    app.shutdown_lsps()
        return result

    def cleanup(self, signum: int, frame: FrameType | None) -> None:
        """Restore the state of the terminal on unexpected exit."""
        log.critical("Unexpected exit signal, restoring terminal")
        output = self.output
        if self.is_running:
            self.exit()
        self.shutdown_lsps()
        # Reset terminal state
        output.reset_cursor_key_mode()
        output.enable_autowrap()
        output.clear_title()
        output.show_cursor()
        output.reset_attributes()
        self.renderer.reset()
        # Exit the main thread
        sys.exit(1)

    @abstractmethod
    def load_container(self) -> AnyContainer:
        """Load the root container for this application.

        Returns:
            The root container for this app

        """
        return FloatContainer(
            content=self.layout.container or Window(),
            floats=cast("list[Float]", self.floats),
        )

    @property
    def tab_registry(self) -> list[TabRegistryEntry]:
        """Return the tab registry."""
        from euporie.core.tabs import _TAB_REGISTRY

        return _TAB_REGISTRY

    def get_file_tabs(self, path: Path) -> list[TabRegistryEntry]:
        """Return the tab to use for a file path."""
        path_mime = get_mime(path) or "text/plain"
        log.debug("File %s has mime type: %s", path, path_mime)

        # Use a set to automatically handle duplicates
        tab_options: set[TabRegistryEntry] = set()
        for entry in self.tab_registry:
            for mime_type in entry.mime_types:
                if PurePath(path_mime).match(mime_type):
                    tab_options.add(entry)
            if path.suffix in entry.file_extensions:
                tab_options.add(entry)

        # Sort by weight (TabRegistryEntry.__lt__ handles this)
        return sorted(tab_options, reverse=True)

    def get_file_tab(self, path: Path) -> type[Tab] | None:
        """Return the tab to use for a file path."""
        if tabs := self.get_file_tabs(path):
            return tabs[0].tab_class
        return None

    def get_language_lsps(self, language: str) -> list[LspClient]:
        """Return the approprrate LSP clients for a given language."""
        clients = []
        if self.config.enable_language_servers:
            from shutil import which

            lsps = {**KNOWN_LSP_SERVERS, **self.config.language_servers}
            for name, kwargs in lsps.items():
                if kwargs:
                    client = None
                    if (
                        (
                            not (lsp_langs := kwargs.get("languages"))
                            or language in lsp_langs
                        )
                        and not (client := self.lsp_clients.get(name))
                        and which(kwargs["command"][0])
                    ):
                        client = LspClient(name, **kwargs)
                        self.lsp_clients[name] = client
                        client.start()
                    if client:
                        clients.append(client)
        return clients

    def shutdown_lsps(self) -> None:
        """Shut down all the remaining LSP servers."""
        from concurrent.futures import as_completed

        # Wait for all LSP exit calls to complete
        # The exit calls occur in the LSP event loop thread
        list(as_completed([lsp.exit() for lsp in self.lsp_clients.values()]))

    def open_file(
        self, path: Path, read_only: bool = False, tab_class: type[Tab] | None = None
    ) -> None:
        """Create a tab for a file.

        Args:
            path: The file path of the notebook file to open
            read_only: If true, the file should be opened read_only
            tab_class: The tab type to use to open the file

        """
        ppath = parse_path(path)
        log.info("Opening file %s", path)
        for tab in self.tabs:
            if ppath == getattr(tab, "path", "") and (
                tab_class is None or isinstance(tab, tab_class)
            ):
                log.info("File %s already open, activating", path)
                self.layout.focus(tab)
                break
        else:
            if tab_class is None:
                tab_class = self.get_file_tab(path)
            if tab_class is None:
                log.error("Unable to display file %s", path)
            else:
                tab = tab_class(self, ppath)
                self.add_tab(tab)
                # Ensure the opened tab is focused at app start
                self.focused_element = tab
                # Ensure the newly opened tab is selected
                self.tab_idx = len(self.tabs) - 1

    def open_files(self) -> None:
        """Open the files defined in the configuration."""
        for file in self.config.files:
            self.open_file(file)

    @property
    def tab(self) -> Tab | None:
        """Return the currently selected tab container object."""
        if self.tabs:
            # Detect if focused tab has changed
            # Find index of selected child
            for i, tab in enumerate(self.tabs):
                if self.render_counter > 0 and self.layout.has_focus(tab):
                    self._tab_idx = i
                    break
            self._tab_idx = max(0, min(self._tab_idx, len(self.tabs) - 1))
            return self.tabs[self._tab_idx]
        else:
            return None

    @property
    def tab_idx(self) -> int:
        """Get the current tab index."""
        return self._tab_idx

    @tab_idx.setter
    def tab_idx(self, value: int) -> None:
        """Set the current tab by index."""
        self._tab_idx = value % (len(self.tabs) or 1)
        if self.tabs:
            container = to_container(self.tabs[self._tab_idx])
            try:
                self.layout.focus(container)
            except ValueError:
                log.exception("Cannot focus tab")

    def focus_tab(self, tab: Tab) -> None:
        """Make a tab visible and focuses it."""
        self.tab_idx = self.tabs.index(tab)

    def cleanup_closed_tab(self, tab: Tab) -> None:
        """Remove a tab container from the current instance of the app.

        Args:
            tab: The closed instance of the tab container

        """
        # Remove tab
        if tab in self.tabs:
            self.tabs.remove(tab)
            self.on_tabs_change()
        # Focus the next active tab if one exists
        if next_tab := self.tab:
            next_tab.focus()
        # If no tab is open, ensure something is focused
        else:
            try:
                self.layout.focus_next()
            except ValueError:
                pass

    def add_tab(self, tab: Tab) -> None:
        """Add a tab to the current tabs list."""
        self.tabs.append(tab)
        self.on_tabs_change()

    def close_tab(self, tab: Tab | None = None) -> None:
        """Close a notebook tab.

        Args:
            tab: The instance of the tab to close. If `None`, the currently
                selected tab will be closed.

        """
        if tab is None:
            tab = self.tab
        if tab is not None:
            tab.close(cb=partial(self.cleanup_closed_tab, tab))

    def get_edit_mode(self) -> EditingMode:
        """Return the editing mode enum defined in the configuration."""
        micro_mode = cast("EditingMode", ExtraEditingMode.MICRO)

        return {
            "micro": micro_mode,
            "vi": EditingMode.VI,
            "emacs": EditingMode.EMACS,
        }.get(str(self.config.edit_mode), micro_mode)

    def update_edit_mode(self, setting: Setting | None = None) -> None:
        """Set the keybindings for editing mode."""
        self.editing_mode = self.get_edit_mode()
        log.debug("Editing mode set to: %s", self.editing_mode)

    @property
    def color_depth(self) -> ColorDepth:
        """The active :class:`.ColorDepth`.

        The current value is determined as follows:

        - If a color depth was given explicitly to this application, use that
          value.
        - Otherwise, fall back to the color depth that is reported by the
          :class:`.Output` implementation. If the :class:`.Output` class was
          created using `output.defaults.create_output`, then this value is
          coming from the $PROMPT_TOOLKIT_COLOR_DEPTH environment variable.
        """
        # Detect terminal color depth
        if self._color_depth is None:
            if os.environ.get("NO_COLOR", "") or os.environ.get("TERM", "") == "dumb":
                self._color_depth = ColorDepth.DEPTH_1_BIT
            colorterm = os.environ.get("COLORTERM", "")
            if "truecolor" in colorterm or "24bit" in colorterm:
                self._color_depth = ColorDepth.DEPTH_24_BIT
            elif "256" in os.environ.get("TERM", ""):
                self._color_depth = ColorDepth.DEPTH_8_BIT

        return super().color_depth

    @property
    def syntax_theme(self) -> str:
        """Calculate the current syntax theme."""
        syntax_theme = self.config.syntax_theme
        if syntax_theme == self.config.defaults.syntax_theme:
            syntax_theme = "tango" if self.color_palette.bg.is_light else "euporie"
        return syntax_theme

    base_styles = (
        Style(MIME_STYLE),
        Style(HTML_STYLE),
        Style(LOG_STYLE),
        Style(IPYWIDGET_STYLE),
    )

    def create_merged_style(self) -> BaseStyle:
        """Generate a new merged style for the application.

        Using a dynamic style has serious performance issues, so instead we update
        the style on the renderer directly when it changes in `self.update_style`

        Returns:
            Return a combined style to use for the application

        """
        styles: list[BaseStyle] = [
            style_from_pygments_cls(get_style_by_name(self.syntax_theme)),
            *self.base_styles,
        ]

        # Get foreground and background colors based on the configured colour scheme
        theme_colors: dict[str, dict[str, str]] = {
            "default": {},
            "light": {"fg": "#202020", "bg": "#F0F0F0"},
            "dark": {"fg": "#F0F0F0", "bg": "#202020"},
            "white": {"fg": "#000000", "bg": "#FFFFFF"},
            "black": {"fg": "#FFFFFF", "bg": "#000000"},
            # TODO - use config.custom_colors
            "custom": {
                "fg": self.config.custom_foreground_color,
                "bg": self.config.custom_background_color,
            },
        }
        base_colors: dict[str, str] = {
            **DEFAULT_COLORS,
            **self.term_colors,
            **theme_colors.get(self.config.color_scheme, theme_colors["default"]),
        }

        # Build a color palette from the fg/bg colors
        self.color_palette = cp = ColorPalette()
        for name, color in base_colors.items():
            cp.add_color(
                name,
                color or theme_colors["default"][name],
                "default" if name in ("fg", "bg") else name,
            )
        # Add accent color
        # self.color_palette.add_color(
        #     "hl",
        #     (bg := self.color_palette.bg)
        #     .adjust(
        #         hue=(bg.hue + (bg.hue - 0.036)) % 1,
        #         saturation=(0.88 - bg.saturation),
        #         brightness=0.4255 - bg.brightness,
        #     )
        #     .base_hex,
        # )
        cp.add_color(
            "hl", base_colors.get(self.config.accent_color, self.config.accent_color)
        )

        # Build app style
        styles.append(build_style(cp))

        # Apply style transformations based on the configured color scheme
        self.style_transformation = merge_style_transformations(
            [
                ConditionalStyleTransformation(
                    SetDefaultColorStyleTransformation(
                        fg=base_colors["fg"], bg=base_colors["bg"]
                    ),
                    self.config.color_scheme != "default",
                ),
                ConditionalStyleTransformation(
                    SwapLightAndDarkStyleTransformation(),
                    self.config.color_scheme == "inverse",
                ),
            ]
        )

        # Add user style customizations
        if custom_style_dict := self.config.custom_styles:
            styles.append(Style.from_dict(custom_style_dict))

        return merge_styles(styles)

    def update_style(self, query: Setting | None = None) -> None:
        """Update the application's style when the syntax theme is changed."""
        self.renderer.style = self.create_merged_style()

    def refresh(self) -> None:
        """Reset all tabs."""
        for tab in self.tabs:
            to_container(tab).reset()

    def _create_merged_style(
        self, include_default_pygments_style: Filter | None = None
    ) -> BaseStyle:
        """Block default style loading."""
        return DummyStyle()

    def draw(self, render_as_done: bool = True) -> None:
        """Draw the app without focus, leaving the cursor below the drawn output."""
        # Hide ephemeral containers
        self._redrawing = True
        # Ensure nothing in the layout has focus
        self.layout._stack.append(Window())
        # Re-draw the app
        self._redraw(render_as_done=render_as_done)
        # Remove the focus block
        self.layout._stack.pop()
        # Show ephemeral containers
        self._redrawing = False
        # Ensure the renderer knows where the cursor is
        self._request_absolute_cursor_position()

    def _handle_exception(
        self, loop: AbstractEventLoop, context: dict[str, Any]
    ) -> None:
        exception = context.get("exception")
        # Log observed exceptions to the log
        log.exception("An unhandled exception occurred", exc_info=exception)

    # async def cancel_and_wait_for_background_tasks(self) -> "None":
    # """Cancel background tasks, ignoring exceptions."""
    # # try:
    # # await super().cancel_and_wait_for_background_tasks()
    # # except ValueError as e:
    # # for task in self._background_tasks:
    # # print(task)
    # # raise e
    #
    # for task in self._background_tasks:
    # print(task)
    # print(task.get_loop())
    # await asyncio.wait([task])

    # ################################# Key Bindings ##################################

    register_bindings(
        {
            "euporie.core.app.app:BaseApp": {
                "quit": ["c-q", "<sigint>"],
                "close-tab": "c-w",
                "next-tab": "c-pagedown",
                "previous-tab": "c-pageup",
                "focus-next": "tab",
                "focus-previous": "s-tab",
                "clear-screen": "c-l",
            }
        }
    )



"""Contain a tab for displaying files."""

from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING, ClassVar

from prompt_toolkit.layout.containers import HSplit
from prompt_toolkit.layout.dimension import Dimension

from euporie.core.filters import insert_mode, replace_mode
from euporie.core.kernel.jupyter import JupyterKernel, MsgCallbacks
from euporie.core.key_binding.registry import load_registered_bindings
from euporie.core.lexers import detect_lexer
from euporie.core.tabs.kernel import KernelTab
from euporie.core.widgets.inputs import KernelInput

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Callable

    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.core.app.app import BaseApp
    from euporie.core.bars.status import StatusBarFields
    from euporie.core.comm.base import Comm

log = logging.getLogger(__name__)


class EditorTab(KernelTab):
    """Tab class for editing text files."""

    name = "Text Editor"
    weight = 1
    mime_types: ClassVar[set[str]] = {"text/*"}
    allow_stdin = True

    def __init__(
        self,
        app: BaseApp,
        path: Path | None = None,
        kernel: JupyterKernel | None = None,
        comms: dict[str, Comm] | None = None,
        use_kernel_history: bool = False,
    ) -> None:
        """Call when the tab is created."""
        self.default_callbacks = MsgCallbacks({})
        self._metadata = {}
        self.loaded = False

        super().__init__(app, path, kernel, comms, use_kernel_history)

    def post_init_kernel(self) -> None:
        """Load UI and file in background after kernel has inited."""
        # Load the UI
        self.container = self.load_container()
        self.app.layout.focus(self.container)
        # Read file
        self.load()
        # Continue loading tab
        super().post_init_kernel()

    def load(self) -> None:
        """Load the text file."""
        try:
            text = self.path.read_text() if self.path is not None else ""
        except FileNotFoundError:
            text = ""

        # Set text
        self.input_box.text = text
        self.input_box.buffer.on_text_changed += lambda b: setattr(self, "dirty", True)
        self.input_box.read_only = False
        self.loaded = True

        # Detect language
        lexer = detect_lexer(text[:1000], self.path)
        if lexer:
            self._metadata = {"kernelspec": {"language": lexer.name.casefold()}}
        # Re-lex the file
        self.input_box.control._fragment_cache.clear()
        self.app.invalidate()

    def close(self, cb: Callable | None = None) -> None:
        """Check if the user want to save an unsaved notebook, then close the file.

        Args:
            cb: A callback to run if after closing the notebook.

        """
        if self.dirty and (unsaved := self.app.dialogs.get("unsaved")):
            unsaved.show(
                tab=self,
                cb=partial(super().close, cb),
            )
        else:
            super().close(cb)

    @property
    def position(self) -> str:
        """Return the position of the cursor in the document."""
        doc = self.input_box.buffer.document
        return f"{doc.cursor_position_row + 1}:{doc.cursor_position_col}"

    def __pt_status__(self) -> StatusBarFields | None:
        """Return a list of statusbar field values shown then this tab is active."""
        if not self.loaded:
            return (["Loading…"], [])
        return (
            [
                ("I" if insert_mode() else ("o" if replace_mode() else ">")),
            ],
            [self.position, str(self.path)],
        )

    @property
    def path_name(self) -> str:
        """Return the path name."""
        if self.path is not None:
            return self.path.name
        else:
            return "(New file)"

    @property
    def title(self) -> str:
        """Return the tab title."""
        return ("* " if self.dirty else "") + self.path_name

    def load_container(self) -> AnyContainer:
        """Load the "tab"'s main container."""
        assert self.path is not None

        self.input_box = self._current_input = KernelInput(
            kernel_tab=self,
            completer=self.completer,
            read_only=True,
            name="code",
            formatters=self.formatters,
            language=lambda: self.language,
            inspector=self.inspector,
            on_text_changed=lambda buf: self.on_change(),
            diagnostics=self.report,
        )

        return HSplit(
            [self.input_box],
            width=Dimension(weight=1),
            height=Dimension(weight=1),
            key_bindings=load_registered_bindings(
                "euporie.core.tabs.base.Tab",
                config=self.app.config,
            ),
        )

    def write_file(self, path: Path) -> None:
        """Write the file's text data to a path.

        Args:
            path: An path at which to save the file

        """
        path.write_text(self.input_box.buffer.text)



"""Responsible for loading data from urls."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

import upath
from aiohttp.client_reqrep import ClientResponse
from fsspec.implementations.http import HTTPFileSystem as FsHTTPFileSystem
from fsspec.implementations.http import get_client
from fsspec.registry import register_implementation as fs_register_implementation
from upath import UPath

if TYPE_CHECKING:
    import asyncio
    from collections.abc import Mapping
    from os import PathLike
    from typing import Any, Callable

    import aiohttp


log = logging.getLogger(__name__)


class NoRaiseClientResponse(ClientResponse):
    """An ``aiohttp`` client response which does not raise on >=400 status responses."""

    @property
    def ok(self) -> bool:
        """Returns ``True`` if ``status`` is probably renderable."""
        return self.status not in {405}


class HTTPFileSystem(FsHTTPFileSystem):
    """A HTTP filesystem implementation which does not raise on errors."""

    def __init__(
        self,
        simple_links: bool = True,
        block_size: int | None = None,
        same_scheme: bool = True,
        size_policy: None = None,
        cache_type: str = "bytes",
        cache_options: dict[str, Any] | None = None,
        asynchronous: bool = False,
        loop: asyncio.AbstractEventLoop | None = None,
        client_kwargs: dict[str, Any] | None = None,
        get_client: Callable[..., aiohttp.ClientSession] = get_client,
        encoded: bool = False,
        **storage_options: Any,
    ) -> None:
        """Defaults to using :py:mod:`NoRaiseClientResponse` for responses."""
        client_kwargs = {
            "response_class": NoRaiseClientResponse,
            **(client_kwargs or {}),
        }
        super().__init__(
            simple_links=simple_links,
            block_size=block_size,
            same_scheme=same_scheme,
            size_policy=size_policy,
            cache_type=cache_type,
            cache_options=cache_options,
            asynchronous=asynchronous,
            loop=loop,
            client_kwargs=client_kwargs,
            get_client=get_client,
            encoded=encoded,
            **storage_options,
        )

    def _raise_not_found_for_status(self, response: ClientResponse, url: str) -> None:
        """Do not raise an exception for 404 errors."""
        response.raise_for_status()


fs_register_implementation("http", HTTPFileSystem, clobber=True)
fs_register_implementation("https", HTTPFileSystem, clobber=True)


# Define custom universal_pathlib path implementations


class UntitledPath(upath.core.UPath):
    """A path for untitled files, as needed for LSP servers."""

    @classmethod
    def _parse_storage_options(
        cls, urlpath: str, protocol: str, storage_options: Mapping[str, Any]
    ) -> dict[str, Any]:
        """Parse storage_options from the urlpath."""
        return {}

    def exists(self, *, follow_symlinks: bool = True) -> bool:
        """Untitled files are unsaved and do not exist."""
        return False


def parse_path(path: str | PathLike, resolve: bool | None = None) -> Path:
    """Parse and resolve a path."""
    if resolve is None:
        from upath.implementations.http import HTTPPath

        if isinstance(path, HTTPPath):
            resolve = True
        else:
            resolve = False
    if not isinstance(path, Path):
        path = UPath(path)
    try:
        path = path.expanduser()
    except NotImplementedError:
        pass
    if resolve:
        try:
            path = path.resolve()
        except (AttributeError, NotImplementedError, Exception):
            log.info("Path %s not resolvable", path)
    else:
        try:
            path = path.absolute()
        except NotImplementedError:
            pass
    return path
